import torch
import numpy as np
from scipy import sparse
from scipy.spatial import distance_matrix
from scipy.special import wrightomega
import matplotlib.pyplot as plt

try:
    from tqdm import tqdm
    HAS_TQDM = True
except ImportError:
    HAS_TQDM = False
    # Fallback progress bar class
    class tqdm:
        def __init__(self, iterable=None, total=None, desc=None, **kwargs):
            self.total = total
            self.desc = desc
            self.n = 0
            
        def __enter__(self):
            if self.desc:
                print(f"{self.desc}: 0/{self.total}")
            return self
            
        def __exit__(self, *args):
            if self.desc:
                print(f"{self.desc}: {self.total}/{self.total} - Complete!")
                
        def update(self, n=1):
            self.n += n
            if self.desc and self.n % max(1, self.total // 10) == 0:
                print(f"{self.desc}: {self.n}/{self.total}")
                
        def set_postfix(self, **kwargs):
            pass


def unot_torch(a,
               b,
               C,
               eps_p,
               rho,
               eps_mu=None,
               eps_nu=None,
               sparse_mtx=False,
               solver="sinkhorn",
               nitermax=10000,
               stopthr=1e-8,
               verbose=False,
               show_progress=False,
               momentum_dt=0.1,
               momentum_beta=0.0,
               device='cuda' if torch.cuda.is_available() else 'cpu'):
    """ The main function calling different algorithms using PyTorch for acceleration.

    Parameters
    ----------
    a : (ns,) numpy.ndarray or torch.Tensor
        Source distribution. The summation should be less than or equal to 1.
    b : (nt,) numpy.ndarray or torch.Tensor
        Target distribution. The summation should be less than or equal to 1.
    C : (ns,nt) numpy.ndarray or torch.Tensor
        The cost matrix possibly with infinity entries.
    eps_p :  float
        The coefficient of entropy regularization for P.
    rho : float
        The coefficient of penalty for unmatched mass.
    eps_mu : float, defaults to eps_p
        The coefficient of entropy regularization for mu.
    eps_nu : float, defaults to eps_p
        The coefficient of entropy regularization for nu.
    sparse_mtx : boolean, defaults to False
        Whether using sparse matrix format. If True, C should be in coo_sparse format.
    solver : str, defaults to 'sinkhorn'
        The solver to use. Choose from 'sinkhorn' and 'momentum'.
    nitermax : int, defaults to 10000
        The maximum number of iterations.
    stopthr : float, defaults to 1e-8
        The relative error threshold for stopping.
    verbose : boolean, defaults to False
        Whether to print algorithm logs.
    show_progress : boolean, defaults to True
        Whether to show progress bar during iterations.
    momentum_dt : float, defaults to 1e-1
        Step size if momentum method is used.
    momentum_beta : float, defautls to 0
        The coefficient for the momentum term if momemtum method is used.
    device : str, defaults to 'cuda' if available else 'cpu'
        Device to run computation on.
    """
    if eps_mu is None: eps_mu = eps_p
    if eps_nu is None: eps_nu = eps_p
    
    # Convert numpy arrays to torch tensors
    if isinstance(a, np.ndarray):
        a = torch.from_numpy(a).float().to(device)
    else:
        a = a.float().to(device)
    
    if isinstance(b, np.ndarray):
        b = torch.from_numpy(b).float().to(device)
    else:
        b = b.float().to(device)
    
    if isinstance(C, np.ndarray):
        C = torch.from_numpy(C).float().to(device)
    elif not isinstance(C, torch.Tensor) and not sparse_mtx:
        C = torch.tensor(C).float().to(device)
    
    # Return a zero matrix if either a or b is all zero
    nzind_a = torch.where(a > 0)[0]
    nzind_b = torch.where(b > 0)[0]
    if len(nzind_a) == 0 or len(nzind_b) == 0:
        if sparse_mtx:
            # Convert back to scipy sparse for consistency
            P = sparse.coo_matrix(([], ([], [])), shape=(len(a), len(b)))
        else:
            P = torch.zeros([len(a), len(b)], dtype=torch.float32, device=device)
        return P
        
    if solver == "sinkhorn" and max(abs(eps_p-eps_mu), abs(eps_p-eps_nu)) > 1e-8:
        print("To use Sinkhorn algorithm, set eps_p=eps_mu=eps_nu")
        return None
        
    if solver == "sinkhorn" and not sparse_mtx:
        P = unot_sinkhorn_l1_dense_torch(a, b, C, eps_p, rho, 
                                         nitermax=nitermax, stopthr=stopthr, 
                                         verbose=verbose, show_progress=show_progress, device=device)
    elif solver == "sinkhorn" and sparse_mtx:
        P = unot_sinkhorn_l1_sparse_torch(a, b, C, eps_p, rho, 
                                          nitermax=nitermax, stopthr=stopthr, 
                                          verbose=verbose, show_progress=show_progress, device=device)
    elif solver == "momentum" and not sparse_mtx: 
        P = unot_momentum_l1_dense_torch(a, b, C, eps_p, eps_mu, eps_nu, rho, 
                                         nitermax=nitermax, stopthr=stopthr, dt=momentum_dt, 
                                         beta=momentum_beta, precondition=True, 
                                         verbose=verbose, show_progress=show_progress, device=device)
    elif solver == "momentum" and sparse_mtx:
        print("Sparse momentum solver under construction")
        return None
    
    return P


def wrightomega_torch(z, device='cpu'):
    """PyTorch implementation of Wright Omega function using scipy fallback"""
    # Convert to numpy for scipy computation, then back to torch
    if isinstance(z, torch.Tensor):
        z_np = z.detach().cpu().numpy()
    else:
        z_np = z
    
    result_np = wrightomega(z_np).real
    return torch.from_numpy(result_np).float().to(device)


def unot_sinkhorn_l1_dense_torch(a, b, C, eps, m, nitermax=10000, stopthr=1e-8, 
                                 verbose=False, output_fg=False, show_progress=True, device='cpu'):
    """ Solve the unnormalized optimal transport with l1 penalty using PyTorch.

    Parameters
    ----------
    a : torch.Tensor
        Source distribution. The summation should be less than or equal to 1.
    b : torch.Tensor
        Target distribution. The summation should be less than or equal to 1.
    C : torch.Tensor
        The cost matrix possibly with infinity entries.
    eps : float
        The coefficient for entropy regularization.
    m : float
        The coefficient for penalizing unmatched mass.
    nitermax : int, optional
        The max number of iterations. Defaults to 10000.
    stopthr : float, optional
        The threshold for terminating the iteration. Defaults to 1e-8.
    verbose : bool, optional
        Whether to print verbose output.
    show_progress : bool, optional
        Whether to show progress bar.
    device : str
        Device to run computation on.

    Returns
    -------
    torch.Tensor
        The optimal transport matrix.
    """
    f = torch.zeros_like(a)
    g = torch.zeros_like(b)
    niter = 0
    err = 100
    
    # Setup progress bar
    pbar = None
    if show_progress:
        pbar = tqdm(total=nitermax, desc="Sinkhorn L1 Dense", 
                   bar_format='{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}] err={postfix}')
    
    try:
        while niter <= nitermax and err > stopthr:
            fprev = f.clone()
            gprev = g.clone()
            
            # Iteration
            # f update
            exp_term = torch.exp((f.unsqueeze(1) + g.unsqueeze(0) - C) / eps)
            sum_exp = torch.sum(exp_term, dim=1)
            exp_f_term = torch.exp((-m + f) / eps)
            
            f = eps * torch.log(a) - eps * torch.log(sum_exp + exp_f_term) + f
            
            # g update
            exp_term = torch.exp((f.unsqueeze(1) + g.unsqueeze(0) - C) / eps)
            sum_exp = torch.sum(exp_term, dim=0)
            exp_g_term = torch.exp((-m + g) / eps)
            
            g = eps * torch.log(b) - eps * torch.log(sum_exp + exp_g_term) + g
            
            # Check relative error
            if niter % 10 == 0:
                err_f = torch.abs(f - fprev).max() / max(torch.abs(f).max(), torch.abs(fprev).max(), 1.)
                err_g = torch.abs(g - gprev).max() / max(torch.abs(g).max(), torch.abs(gprev).max(), 1.)
                err = 0.5 * (err_f + err_g)
                
                # Update progress bar
                if pbar is not None:
                    pbar.set_postfix_str(f"{err:.2e}")
                    
            niter = niter + 1
            
            # Update progress bar
            if pbar is not None:
                pbar.update(1)
                
    finally:
        if pbar is not None:
            pbar.close()
    
    if verbose:
        print('Number of iterations in unot:', niter)
    
    P = torch.exp((f.unsqueeze(1) + g.unsqueeze(0) - C) / eps)
    
    if output_fg:
        return f, g
    else:
        return P


def unot_sinkhorn_l1_sparse_torch(a, b, C, eps, m, nitermax=10000, stopthr=1e-8, 
                                  verbose=False, show_progress=True, device='cpu'):
    """ Solve the unnormalized optimal transport with l1 penalty using PyTorch for sparse matrices.

    Parameters
    ----------
    a : torch.Tensor
        Source distribution. The summation should be less than or equal to 1.
    b : torch.Tensor
        Target distribution. The summation should be less than or equal to 1.
    C : scipy.sparse.coo_matrix
        The cost matrix in coo sparse format.
    eps : float
        The coefficient for entropy regularization.
    m : float
        The coefficient for penalizing unmatched mass.
    nitermax : int, optional
        The max number of iterations. Defaults to 10000.
    stopthr : float, optional
        The threshold for terminating the iteration. Defaults to 1e-8.
    verbose : bool, optional
        Whether to print verbose output.
    show_progress : bool, optional
        Whether to show progress bar.
    device : str
        Device to run computation on.

    Returns
    -------
    scipy.sparse.coo_matrix
        The optimal transport matrix.
    """
    # Convert sparse matrix indices and data to torch tensors
    row_idx = torch.from_numpy(C.row).long().to(device)
    col_idx = torch.from_numpy(C.col).long().to(device)
    C_data = torch.from_numpy(C.data).float().to(device)
    
    f = torch.zeros_like(a)
    g = torch.zeros_like(b)
    niter = 0
    err = 100
    
    # Setup progress bar
    pbar = None
    if show_progress:
        pbar = tqdm(total=nitermax, desc="Sinkhorn L1 Sparse", 
                   bar_format='{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}] err={postfix}')
    
    try:
        while niter <= nitermax and err > stopthr:
            fprev = f.clone()
            gprev = g.clone()
            
            # Compute sparse exponential terms
            exp_data = torch.exp((-C_data + f[row_idx] + g[col_idx]) / eps)
            
            # f update
            sum_f = torch.zeros_like(a)
            sum_f.scatter_add_(0, row_idx, exp_data)
            exp_f_term = torch.exp((-m + f) / eps)
            f = eps * torch.log(a) - eps * torch.log(sum_f + exp_f_term) + f
            
            # Recompute exp_data for g update
            exp_data = torch.exp((-C_data + f[row_idx] + g[col_idx]) / eps)
            
            # g update
            sum_g = torch.zeros_like(b)
            sum_g.scatter_add_(0, col_idx, exp_data)
            exp_g_term = torch.exp((-m + g) / eps)
            g = eps * torch.log(b) - eps * torch.log(sum_g + exp_g_term) + g
            
            # Check relative error
            if niter % 10 == 0:
                err_f = torch.abs(f - fprev).max() / max(torch.abs(f).max(), torch.abs(fprev).max(), 1.)
                err_g = torch.abs(g - gprev).max() / max(torch.abs(g).max(), torch.abs(gprev).max(), 1.)
                err = 0.5 * (err_f + err_g)
                
                # Update progress bar
                if pbar is not None:
                    pbar.set_postfix_str(f"{err:.2e}")
                    
            niter = niter + 1
            
            # Update progress bar
            if pbar is not None:
                pbar.update(1)
                
    finally:
        if pbar is not None:
            pbar.close()

    if verbose:
        print('Number of iterations in unot:', niter)
    
    # Create final result as sparse matrix
    final_exp_data = torch.exp((-C_data + f[row_idx] + g[col_idx]) / eps)
    
    # Convert back to scipy sparse matrix
    result_data = final_exp_data.detach().cpu().numpy()
    result_row = row_idx.detach().cpu().numpy()
    result_col = col_idx.detach().cpu().numpy()
    
    result_sparse = sparse.coo_matrix((result_data, (result_row, result_col)), 
                                      shape=(len(a), len(b)))
    
    return result_sparse


def unot_momentum_l1_dense_torch(a, b, C, eps_p, eps_mu, eps_nu, m, nitermax=1e4, 
                                 stopthr=1e-8, dt=0.01, beta=0.8, precondition=False, 
                                 verbose=False, show_progress=True, device='cpu'):
    """ Solve unnormalized optimal transport using momentum method with PyTorch.

    Parameters
    ----------
    a : torch.Tensor
        Source distribution.
    b : torch.Tensor
        Target distribution.
    C : torch.Tensor
        Cost matrix.
    eps_p : float
        Entropy regularization coefficient for P.
    eps_mu : float
        Entropy regularization coefficient for mu.
    eps_nu : float
        Entropy regularization coefficient for nu.
    m : float
        Penalty coefficient for unmatched mass.
    nitermax : float, optional
        Maximum number of iterations.
    stopthr : float, optional
        Stopping threshold.
    dt : float, optional
        Step size.
    beta : float, optional
        Momentum coefficient.
    precondition : bool, optional
        Whether to use preconditioning.
    verbose : bool, optional
        Whether to print verbose output.
    show_progress : bool, optional
        Whether to show progress bar.
    device : str
        Device to run computation on.

    Returns
    -------
    torch.Tensor
        The optimal transport matrix.
    """
    f = torch.zeros_like(a)
    g = torch.zeros_like(b)
    
    if precondition:
        f, g = unot_sinkhorn_l1_dense_torch(a, b, C, eps_p, m, output_fg=True, 
                                           show_progress=show_progress, device=device)
    
    f_old = f.clone()
    g_old = g.clone()
    F_old = f.clone()
    G_old = g.clone()
    niter = 0
    err = 100
    
    def Qf(ff, gg, ee_p, ee_mu, ee_nu, mm, aa, bb, CC):
        exp_term = torch.exp(ff / ee_p) * torch.sum(torch.exp((gg.unsqueeze(0) - CC) / ee_p), dim=1)
        penalty_term = torch.exp((ff - mm) / ee_mu)
        return exp_term + penalty_term - aa
    
    def Qg(ff, gg, ee_p, ee_mu, ee_nu, mm, aa, bb, CC):
        exp_term = torch.exp(gg / ee_p) * torch.sum(torch.exp((ff.unsqueeze(1) - CC) / ee_p), dim=0)
        penalty_term = torch.exp((gg - mm) / ee_nu)
        return exp_term + penalty_term - bb
    
    nitermax = int(nitermax)
    
    # Setup progress bar
    pbar = None
    if show_progress:
        pbar = tqdm(total=nitermax, desc="Momentum L1 Dense", 
                   bar_format='{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}] err={postfix}')
    
    try:
        while niter <= nitermax and err > stopthr:
            F = beta * F_old + Qf(f_old, g_old, eps_p, eps_mu, eps_nu, m, a, b, C)
            f = f_old - dt * F
            G = beta * G_old + Qg(f_old, g_old, eps_p, eps_mu, eps_nu, m, a, b, C)
            g = g_old - dt * G
            
            if niter % 10 == 0:
                err_f = torch.abs(f - f_old).max() / max(torch.abs(f).max(), torch.abs(f_old).max(), 1.)
                err_g = torch.abs(g - g_old).max() / max(torch.abs(g).max(), torch.abs(g_old).max(), 1.)
                err = 0.5 * (err_f + err_g)
                
                # Update progress bar
                if pbar is not None:
                    pbar.set_postfix_str(f"{err:.2e}")
                    
            f_old = f.clone()
            F_old = F.clone()
            g_old = g.clone()
            G_old = G.clone()
            niter += 1
            
            # Update progress bar
            if pbar is not None:
                pbar.update(1)
                
    finally:
        if pbar is not None:
            pbar.close()
    
    P = torch.exp((f.unsqueeze(1) + g.unsqueeze(0) - C) / eps_p)
    
    if verbose:
        print(f'Number of iterations: {niter}')
    
    return P


def unot_sinkhorn_l2_dense_torch(a, b, C, eps, m, nitermax=10000, stopthr=1e-8, 
                                 verbose=False, show_progress=True, device='cpu'):
    """ Solve the unnormalized optimal transport with l2 penalty using PyTorch.

    Parameters
    ----------
    a : torch.Tensor
        Source distribution.
    b : torch.Tensor
        Target distribution.
    C : torch.Tensor
        The cost matrix.
    eps : float
        The coefficient for entropy regularization.
    m : float
        The coefficient for penalizing unmatched mass.
    nitermax : int, optional
        The max number of iterations.
    stopthr : float, optional
        The threshold for terminating the iteration.
    verbose : bool, optional
        Whether to print verbose output.
    show_progress : bool, optional
        Whether to show progress bar.
    device : str
        Device to run computation on.

    Returns
    -------
    torch.Tensor
        The optimal transport matrix.
    """
    f = torch.zeros_like(a)
    g = torch.zeros_like(b)
    r = torch.zeros_like(a)
    s = torch.zeros_like(b)
    niter = 0
    err = 100
    
    # Setup progress bar
    pbar = None
    if show_progress:
        pbar = tqdm(total=nitermax, desc="Sinkhorn L2 Dense", 
                   bar_format='{desc}: {percentage:3.0f}%|{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}] err={postfix}')
    
    try:
        while niter <= nitermax and err > stopthr:
            fprev = f.clone()
            gprev = g.clone()
            
            # Iteration
            exp_term = torch.exp((f.unsqueeze(1) + g.unsqueeze(0) - C) / eps)
            sum_exp_f = torch.sum(exp_term, dim=1)
            exp_r_term = torch.exp((r + f) / eps)
            
            f = eps * torch.log(a) - eps * torch.log(sum_exp_f + exp_r_term) + f
            
            exp_term = torch.exp((f.unsqueeze(1) + g.unsqueeze(0) - C) / eps)
            sum_exp_g = torch.sum(exp_term, dim=0)
            exp_s_term = torch.exp((s + g) / eps)
            
            g = eps * torch.log(b) - eps * torch.log(sum_exp_g + exp_s_term) + g
            
            # Update r and s using wright omega (fallback to cpu for scipy)
            r = -eps * wrightomega_torch(f / eps - torch.log(torch.tensor(eps / m)), device)
            s = -eps * wrightomega_torch(g / eps - torch.log(torch.tensor(eps / m)), device)
            
            # Check relative error
            if niter % 10 == 0:
                err_f = torch.abs(f - fprev).max() / max(torch.abs(f).max(), torch.abs(fprev).max(), 1.)
                err_g = torch.abs(g - gprev).max() / max(torch.abs(g).max(), torch.abs(gprev).max(), 1.)
                err = 0.5 * (err_f + err_g)
                
                # Update progress bar
                if pbar is not None:
                    pbar.set_postfix_str(f"{err:.2e}")
                    
            niter = niter + 1
            
            # Update progress bar
            if pbar is not None:
                pbar.update(1)
                
    finally:
        if pbar is not None:
            pbar.close()
    
    if verbose:
        print('Number of iterations in unot:', niter)
    
    P = torch.exp((f.unsqueeze(1) + g.unsqueeze(0) - C) / eps)
    return P


# Utility function to convert between numpy and torch
def to_numpy(tensor):
    """Convert torch tensor to numpy array"""
    if isinstance(tensor, torch.Tensor):
        return tensor.detach().cpu().numpy()
    return tensor


def to_torch(array, device='cpu'):
    """Convert numpy array to torch tensor"""
    if isinstance(array, np.ndarray):
        return torch.from_numpy(array).float().to(device)
    elif isinstance(array, torch.Tensor):
        return array.float().to(device)
    else:
        return torch.tensor(array).float().to(device) 


r"""
Graph and GAN networks in SLAT
"""
import math
from typing import Optional, List
import torch
import torch.nn.functional as F

from .graphconv import CombUnweighted

        
class LGCN(torch.nn.Module):
    r"""
    Lightweight GCN which remove nonlinear functions and concatenate the embeddings of each layer:
        (:math:`Z = f_{e}(A, X) = Concat( [X, A_{X}, A_{2X}, ..., A_{KX}])W_{e}`)
    
    Parameters
    ----------
    input_size
        input dim
    K
        LGCN layers
    """
    def __init__(self, input_size:int, K:Optional[int]=8):
        super(LGCN, self).__init__()
        self.conv1 = CombUnweighted(K=K)
        
    def forward(self, feature:torch.Tensor, edge_index:torch.Tensor):
        x = self.conv1(feature, edge_index)
        return x


class LGCN_mlp(torch.nn.Module):
    r"""
    Add one hidden layer MLP in LGCN
    
    Parameters
    ----------
    input_size
        input dim
    output_size
        output dim
    K
        LGCN layers
    hidden_size
        hidden size of MLP
    dropout
        dropout ratio
    """
    def __init__(self, input_size:int, output_size:int, K:Optional[int]=8,
                 hidden_size:Optional[int]=512, dropout:Optional[float]=0.2):
        super(LGCN_mlp, self).__init__()
        self.conv1 = CombUnweighted(K=K)
        self.fc1 = torch.nn.Linear(input_size * (K + 1), hidden_size)
        self.bn = torch.nn.BatchNorm1d(hidden_size)
        self.dropout1 = torch.nn.Dropout(p=dropout)
        self.fc2 = torch.nn.Linear(hidden_size, output_size)
        
    def forward(self, feature:torch.Tensor, edge_index:torch.Tensor):
        x = self.conv1(feature, edge_index)
        x = F.leaky_relu(self.fc1(x), negative_slope=0.2)
        x = self.bn(x)
        x = self.dropout1(x)
        x = self.fc2(x)
        return x


class WDiscriminator(torch.nn.Module):
    r"""
    WGAN Discriminator
    
    Parameters
    ----------
    hidden_size
        input dim
    hidden_size2
        hidden dim
    """
    def __init__(self, hidden_size:int, hidden_size2:Optional[int]=512):
        super(WDiscriminator, self).__init__()
        self.hidden = torch.nn.Linear(hidden_size, hidden_size2)
        self.hidden2 = torch.nn.Linear(hidden_size2, hidden_size2)
        self.output = torch.nn.Linear(hidden_size2, 1)
    def forward(self, input_embd):
        return self.output(F.leaky_relu(self.hidden2(F.leaky_relu(self.hidden(input_embd), 0.2, inplace=True)), 0.2, inplace=True))


class transformation(torch.nn.Module):
    r"""
    Transformation in LGCN
    
    Parameters
    ----------
    hidden_size
        input dim
    """
    def __init__(self, hidden_size:Optional[int]=512):
        super(transformation, self).__init__()
        self.trans = torch.nn.Parameter(torch.eye(hidden_size))
    def forward(self, input_embd):
        return input_embd.mm(self.trans)


class notrans(torch.nn.Module):
    r"""
    LGCN without transformation
    """
    def __init__(self):
        super(notrans, self).__init__()
    def forward(self, input_embd:torch.Tensor):
        return input_embd


class ReconDNN(torch.nn.Module):
    r"""
    Data reconstruction network
    
    Parameters
    ----------
    hidden_size
        input dim
    feature_size
        output size (feature input size)
    hidden_size2
        hidden size
    """
    def __init__(self, hidden_size:int, feature_size:int, hidden_size2:Optional[int]=512):
        super(ReconDNN, self).__init__()
        self.hidden = torch.nn.Linear(hidden_size, hidden_size2)
        self.output = torch.nn.Linear(hidden_size2, feature_size)
    def forward(self, input_embd:torch.Tensor):
        return self.output(F.relu(self.hidden(input_embd)))


r"""Module providing encapsulation of SpaceFlow for spatial flow analysis."""
import random
import numpy as np
import pandas as pd
import scanpy as sc
import torch
from tqdm import tqdm
import torch.nn as nn

import matplotlib.pyplot as plt
from scipy.spatial import distance_matrix
from torch_geometric.nn import GCNConv, DeepGraphInfomax
from sklearn.neighbors import kneighbors_graph
from .._settings import add_reference

sf_install = False

class pySpaceFlow(object):
    r"""SpaceFlow spatial flow analysis class.
    
    SpaceFlow is a deep learning method for analyzing spatial transcriptomics data
    by learning spatially-aware cell representations. It combines graph neural networks
    with spatial regularization to capture both transcriptional and spatial relationships
    between cells.

    The method:
    1. Constructs a spatial neighborhood graph
    2. Learns embeddings using deep graph infomax
    3. Applies spatial regularization to preserve spatial structure
    4. Generates pseudo-spatial maps for trajectory analysis

    Attributes:
        adata: AnnData
            Input annotated data matrix containing:
            - Gene expression data in adata.X
            - Spatial coordinates in adata.obsm['spatial']
        sf: SpaceFlow
            Internal SpaceFlow object for computations
        embedding: array
            Learned spatial-aware embeddings after training

    Examples:
        >>> import scanpy as sc
        >>> import omicverse as ov
        >>> # Load spatial data
        >>> adata = sc.read_visium(...)
        >>> # Initialize SpaceFlow
        >>> spaceflow = ov.space.pySpaceFlow(adata)
        >>> # Train model
        >>> embedding = spaceflow.train(
        ...     spatial_regularization_strength=0.1,
        ...     z_dim=50,
        ...     epochs=1000
        ... )
        >>> # Calculate pseudo-spatial map
        >>> psm = spaceflow.cal_pSM(n_neighbors=20)
    """
    def __init__(self,adata) -> None:
        r"""Initialize SpaceFlow spatial analysis object.
        
        Arguments:
            adata: AnnData
                Annotated data matrix containing:
                - Gene expression data in adata.X
                - Spatial coordinates in adata.obsm['spatial']
                The data should be preprocessed (normalized, scaled)

        Notes:
            - Automatically checks for SpaceFlow package installation
            - Constructs initial spatial neighborhood graph
            - Uses 10 nearest neighbors for graph construction
            - Stores SpaceFlow object in self.sf
        """
        global sf_install
        try:
            from ..externel.spaceflow import SpaceFlow
            sf_install=True
        except ImportError as e:
            raise ImportError(
                'Please install the SpaceFlow: `pip install SpaceFlow`.'
            ) from e
        from ..externel.spaceflow import SpaceFlow
        sf = SpaceFlow(adata=adata, 
                         spatial_locs=adata.obsm['spatial'])

        spatial_locs = adata.obsm['spatial']
        spatial_graph = sf.graph_alpha(spatial_locs, n_neighbors=10)

        sf.adata_preprocessed = adata
        sf.spatial_graph = spatial_graph
        self.sf = sf
        self.adata = adata

    def train(self,spatial_regularization_strength=0.1, 
              z_dim=50, lr=1e-3, epochs=1000, max_patience=50, 
              min_stop=100, random_seed=42, gpu=0, 
              regularization_acceleration=True, edge_subset_sz=1000000):
        r"""Train SpaceFlow model for spatial embedding.
        
        This method trains a graph neural network to learn spatially-aware cell
        representations using deep graph infomax with spatial regularization.

        Arguments:
            spatial_regularization_strength: float, optional (default=0.1)
                Weight for spatial regularization term.
                Higher values enforce stronger spatial consistency.
            z_dim: int, optional (default=50)
                Dimensionality of learned embedding space.
            lr: float, optional (default=1e-3)
                Learning rate for Adam optimizer.
            epochs: int, optional (default=1000)
                Maximum number of training epochs.
            max_patience: int, optional (default=50)
                Number of epochs to wait for improvement before early stopping.
            min_stop: int, optional (default=100)
                Minimum number of epochs before allowing early stopping.
            random_seed: int, optional (default=42)
                Random seed for reproducibility.
            gpu: int, optional (default=0)
                GPU device index to use. Uses CPU if GPU unavailable.
            regularization_acceleration: bool, optional (default=True)
                Whether to use subsampling for faster regularization.
            edge_subset_sz: int, optional (default=1000000)
                Number of edges to sample for accelerated regularization.
            
        Returns:
            numpy.ndarray
                Learned embedding matrix of shape (n_cells, z_dim).

        Notes:
            - Uses deep graph infomax for self-supervised learning
            - Applies spatial regularization to preserve spatial structure
            - Employs early stopping based on loss convergence
            - Supports GPU acceleration when available
            - Results are stored in adata.obsm['spaceflow']
            - Progress is shown with a progress bar
        """
        from ..externel.spaceflow import sparse_mx_to_torch_edge_list, corruption

        adata_preprocessed, spatial_graph = self.sf.adata_preprocessed, self.sf.spatial_graph
        if not adata_preprocessed:
            print("Data has not been preprocessed, please run preprocessing_data() method first!")
            return

        torch.manual_seed(random_seed)
        random.seed(random_seed)
        np.random.seed(random_seed)

        device = f"cuda:{gpu}" if torch.cuda.is_available() else 'cpu'
        model = DeepGraphInfomax(
            hidden_channels=z_dim, encoder=GraphEncoder(adata_preprocessed.shape[1], z_dim),
            summary=lambda z, *args, **kwargs: torch.sigmoid(z.mean(dim=0)),
            corruption=corruption).to(device)

        expr = adata_preprocessed.X.todense() if type(adata_preprocessed.X).__module__ != np.__name__ else adata_preprocessed.X
        expr = torch.tensor(expr).float().to(device)

        edge_list = sparse_mx_to_torch_edge_list(spatial_graph).to(device)

        model.train()
        min_loss = np.inf
        patience = 0
        optimizer = torch.optim.Adam(model.parameters(), lr=lr)
        best_params = model.state_dict()

        for epoch in tqdm(range(epochs)):
            train_loss = 0.0
            torch.set_grad_enabled(True)
            optimizer.zero_grad()
            z, neg_z, summary = model(expr, edge_list)
            loss = model.loss(z, neg_z, summary)

            coords = torch.tensor(adata_preprocessed.obsm['spatial']).float().to(device)
            if regularization_acceleration or adata_preprocessed.shape[0] > 5000:
                cell_random_subset_1, cell_random_subset_2 = torch.randint(0, z.shape[0], (edge_subset_sz,)).to(
                    device), torch.randint(0, z.shape[0], (edge_subset_sz,)).to(device)
                z1, z2 = torch.index_select(z, 0, cell_random_subset_1), torch.index_select(z, 0, cell_random_subset_2)
                c1, c2 = torch.index_select(coords, 0, cell_random_subset_1), torch.index_select(coords, 0,
                                                                                                 cell_random_subset_1)
                pdist = torch.nn.PairwiseDistance(p=2)

                z_dists = pdist(z1, z2)
                z_dists = z_dists / torch.max(z_dists)

                sp_dists = pdist(c1, c2)
                sp_dists = sp_dists / torch.max(sp_dists)
                n_items = z_dists.size(dim=0)
            else:
                z_dists = torch.cdist(z, z, p=2)
                z_dists = torch.div(z_dists, torch.max(z_dists)).to(device)
                sp_dists = torch.cdist(coords, coords, p=2)
                sp_dists = torch.div(sp_dists, torch.max(sp_dists)).to(device)
                n_items = z.size(dim=0) * z.size(dim=0)

            penalty_1 = torch.div(torch.sum(torch.mul(1.0 - z_dists, sp_dists)), n_items).to(device)
            loss = loss + spatial_regularization_strength * penalty_1

            loss.backward()
            optimizer.step()
            train_loss += loss.item()

            if train_loss > min_loss:
                patience += 1
            else:
                patience = 0
                min_loss = train_loss
                best_params = model.state_dict()
            if patience > max_patience and epoch > min_stop:
                break

        model.load_state_dict(best_params)

        z, _, _ = model(expr, edge_list)
        embedding = z.cpu().detach().numpy()

        self.sf.embedding = embedding
        self.adata.obsm['spaceflow']=self.sf.embedding.copy()
        add_reference(self.adata,'SpaceFlow','embedding with SpaceFlow')

        return embedding

    def cal_pSM(self,n_neighbors:int=20,resolution:int=1,
                       max_cell_for_subsampling:int=5000,
                       psm_key='pSM_spaceflow'):
        r"""Calculate pseudo-spatial map using diffusion pseudotime.
        
        This method constructs a pseudo-spatial map by computing diffusion pseudotime
        on the learned embeddings, useful for analyzing spatial trajectories and
        organization patterns.

        Arguments:
            n_neighbors: int, optional (default=20)
                Number of neighbors for kNN graph construction.
                Higher values create denser connectivity.
            resolution: int, optional (default=1)
                Resolution parameter for Leiden clustering.
                Higher values yield more fine-grained clusters.
            max_cell_for_subsampling: int, optional (default=5000)
                Maximum number of cells to use for distance calculations.
                Enables analysis of large datasets through subsampling.
            psm_key: str, optional (default='pSM_spaceflow')
                Key in adata.obs where pseudo-spatial map values will be stored.

        Returns:
            numpy.ndarray
                Pseudo-spatial map values for each cell.

        Notes:
            - Constructs neighborhood graph from embeddings
            - Performs UMAP and Leiden clustering
            - Uses PAGA for trajectory inference
            - Computes diffusion pseudotime
            - Results are stored in adata.obs[psm_key]
            - Useful for ordering cells along spatial axes
            - Handles large datasets through subsampling
        """
        sc.pp.neighbors(self.adata, n_neighbors=n_neighbors,
               use_rep='spaceflow')
        sc.tl.umap(self.adata)
        sc.tl.leiden(self.adata, resolution=resolution)
        sc.tl.paga(self.adata)

        if self.adata.shape[0] < max_cell_for_subsampling:
            sub_adata_x = self.adata.obsm['spaceflow']
        else:
            indices = np.arange(self.adata.shape[0])
            selected_ind = np.random.choice(indices, max_cell_for_subsampling, False)
            sub_adata_x = self.adata[selected_ind, :].obsm['spaceflow']

        sum_dists = distance_matrix(sub_adata_x, sub_adata_x).sum(axis=1)
        self.adata.uns['iroot'] = np.argmax(sum_dists)
        sc.tl.diffmap(self.adata)
        sc.tl.dpt(self.adata)
        self.adata.obs.rename({"dpt_pseudotime": psm_key}, axis=1, inplace=True)
        print(f'The pseudo-spatial map values are stored in adata.obs["{psm_key}"].')
        add_reference(self.adata,'SpaceFlow','pseudo-spatial map with SpaceFlow')
        psm_values = self.adata.obs[psm_key].to_numpy()
        return psm_values


class GraphEncoder(nn.Module):
    r"""Graph convolutional encoder for SpaceFlow.
    
    This class implements a two-layer graph convolutional network (GCN) that serves
    as the encoder component in the SpaceFlow model. It learns to transform gene
    expression data into a spatially-aware latent representation.

    Architecture:
    1. Input layer: Gene expression features
    2. First GCN layer with PReLU activation
    3. Second GCN layer with PReLU activation
    4. Output layer: Learned embeddings

    Attributes:
        conv1: GCNConv
            First graph convolutional layer
        conv2: GCNConv
            Second graph convolutional layer
        prelu: PReLU
            Parametric ReLU activation function

    Notes:
        - Uses PyTorch Geometric's GCNConv implementation
        - Learns spatially-aware representations
        - Preserves both local and global structure
        - Part of the deep graph infomax framework
    """
    def __init__(self, in_channels, hidden_channels):
        super(GraphEncoder, self).__init__()
        self.conv = GCNConv(in_channels, hidden_channels, cached=False)
        self.prelu = nn.PReLU(hidden_channels)
        self.conv2 = GCNConv(hidden_channels, hidden_channels, cached=False)
        self.prelu2 = nn.PReLU(hidden_channels)

    def forward(self, x, edge_index):
        r"""Forward propagation through graph encoder.
        
        Arguments:
            x: Input node features.
            edge_index: Graph edge indices.
            
        Returns:
            Encoded node representations.
        """
        x = self.conv(x, edge_index)
        x = self.prelu(x)
        x = self.conv2(x, edge_index)
        x = self.prelu2(x)
        return x
    # End-of-file (EOF)



import pandas as pd
import numpy as np
import sklearn.neighbors
import scipy.sparse as sp
import seaborn as sns
import matplotlib.pyplot as plt

import scanpy as sc
import torch
from torch_geometric.data import Data
import anndata as ad


def Count2Binary(adata):
    """
    Convert raw count data to binary format (0 or 1) in an AnnData object.

    Parameters:
    adata : AnnData
        The input AnnData object containing the matrix of raw counts. The matrix
        should be accessible via `adata.X`. This matrix should be numerical.

    Returns:
    AnnData
        The modified AnnData object where the original raw count matrix is
        replaced with a binary matrix. In this binary matrix, any non-zero value
        in the original data is set to 1, and zero values remain zero.

    Example:
    >>> adata = AnnData(np.array([[0, 2], [3, 0], [1, 1]]))
    >>> Count2Binary(adata)
    >>> print(adata.X)
    array([[0., 1.],
           [1., 0.],
           [1., 1.]])
    """
    # Convert raw count data to binary by checking if each element is greater than 0.
    # If true, the value becomes 1 (converted to float64). If false, remains 0.
    adata.X = (adata.X > 0).astype(np.float64)
    return adata



def Multi_Refine_label(adata, radius=30, key='mclust', add_key='refine_label', obs_key='slice_id'):
    """
    Refine labels for each unique section in an AnnData object by calling the Refine_label function.

    Parameters:
    adata : AnnData
        The input AnnData object containing the data and labels to be refined.

    radius : int, optional (default=30)
        The number of neighboring cells used to refine the labels in the Refine_label function.

    key : str, optional (default='mclust')
        The key in adata.obs used to access the old labels that need to be refined.

    add_key : str, optional (default='refine_label')
        The key in adata.obs where the refined labels will be stored.

    obs_key : str, optional (default='slice_id')
        The key in adata.obs used to access unique sections.

    Returns:
    AnnData
        The modified AnnData object with refined labels.

    """
    unique_sections = adata.obs[obs_key].unique()
    for section in unique_sections:
        section_data = adata[adata.obs[obs_key] == section].copy()
        refined_labels = Refine_label(section_data, radius=radius, key=key)
        adata.obs.loc[adata.obs[obs_key] == section, add_key] = refined_labels
    return adata


def Refine_label(adata, radius=30, key='label'):
    """
    Refine cell labels based on their spatial neighbors' labels.

    Parameters:
    adata : AnnData
        The input AnnData object containing the data and labels to be refined.

    radius : int, optional (default=30)
        The number of neighboring cells used for refining the label of each cell.

    key : str, optional (default='label')
        The key in adata.obs used to access the old labels that need to be refined.

    Returns:
    list
        A list of refined labels as strings.

    """
    n_neigh = radius
    new_type = []
    old_type = adata.obs[key].values
    
    #calculate distance
    import ot
    position = adata.obsm['spatial']
    distance = ot.dist(position, position, metric='euclidean')
           
    n_cell = distance.shape[0]
    
    for i in range(n_cell):
        vec  = distance[i, :]
        index = vec.argsort()
        neigh_type = []
        for j in range(1, n_neigh+1):
            neigh_type.append(old_type[index[j]])
        max_type = max(neigh_type, key=neigh_type.count)
        new_type.append(max_type)
        
    new_type = [str(i) for i in list(new_type)]    
    return new_type


def find_optimal_resolution(adata, desired_clusters, algorithm="leiden", add_key=None, start=0, end=5, step=1, max_iter=50):
    """
    Find the optimal resolution using a binary search method.
    
    Parameters:
        adata: AnnData object
        desired_clusters: Desired number of clusters
        algorithm: Clustering algorithm, choose between 'leiden' or 'louvain'
        add_key: Key used to store clustering results in adata.obs
        start: Starting resolution
        end: Ending resolution
        step: Step size to adjust resolution
        max_iter: Maximum number of iterations
        
    Returns:
        best_resolution: The optimal resolution
    """
    
    if algorithm not in ["leiden", "louvain"]:
        raise ValueError("Invalid algorithm choice. Choose 'leiden' or 'louvain'")
    
    if add_key is None:
        add_key = algorithm
    
    iteration = 0
    
    while iteration < max_iter:
        midpoint = (start + end) / 2
        
        if algorithm == "leiden":
            sc.tl.leiden(adata, resolution=midpoint, key_added=add_key)
        else:
            sc.tl.louvain(adata, resolution=midpoint, key_added=add_key)
        
        num_clusters = len(np.unique(adata.obs[add_key]))
        
        if num_clusters == desired_clusters:
            print(f"The optimal resolution is: {midpoint}")
            print(f"Stored in adata.obs[{add_key}]:", add_key)
            return midpoint
        elif num_clusters < desired_clusters:
            start = midpoint
        else:
            end = midpoint
            
        step = step / 2
        iteration += 1
        
    print("The optimal resolution was not found within the given number of iterations.")
    print("Try to fine-tune the start and end values.")


def Transfer_pytorch_Data(adata):
    """
    Convert an AnnData object to a PyTorch-Geometric Data object, incorporating spatial relationships.

    Returns:
    
    A PyTorch-Geometric Data object with the edge_index representing the 
    spatial relationships between cells and x representing the expression data.

    """
    G_df = adata.uns['Spatial_Graph'].copy()
    cells = np.array(adata.obs_names)
    cells_id_tran = dict(zip(cells, range(cells.shape[0])))
    G_df['Cell_i'] = G_df['Cell_i'].map(cells_id_tran)
    G_df['Cell_j'] = G_df['Cell_j'].map(cells_id_tran)

    G = sp.coo_matrix((np.ones(G_df.shape[0]), (G_df['Cell_i'], G_df['Cell_j'])), shape=(adata.n_obs, adata.n_obs))
    G = G + sp.eye(G.shape[0])

    edgeList = np.nonzero(G)
    if type(adata.X) == np.ndarray:
        data = Data(edge_index=torch.LongTensor(np.array(
            [edgeList[0], edgeList[1]])), x=torch.FloatTensor(adata.X))  # .todense()
    else:
        data = Data(edge_index=torch.LongTensor(np.array(
            [edgeList[0], edgeList[1]])), x=torch.FloatTensor(adata.X.todense()))  # .todense()
    return data


def Construct_Spatial_Graph(adata, use_method='KNN', cutoff = None):
    """
    Construct a spatial graph based on either K-Nearest Neighbors (KNN) or a fixed radius.

    Parameters:
    adata : AnnData
        The input AnnData object containing spatial coordinates in the obsm['spatial'] attribute.

    use_method : str (default='KNN')
        The method used to construct the spatial graph. 
        Acceptable values: 'KNN' or 'Radius'.

    cutoff : int
        For 'KNN' method, it determines the number of nearest neighbors.
        For 'Radius' method, it determines the distance threshold for considering neighbors.

    Returns:
    None
        Modifies the input adata object to include a 'Spatial_Graph' uns attribute containing the spatial graph.
    """
    
    assert(use_method in ['Radius', 'KNN'])
    print('------Constructing spatial graph...------')
    coor = pd.DataFrame(adata.obsm['spatial'])
    coor.index = adata.obs.index
    coor.columns = ['imagerow', 'imagecol']

    if use_method == 'KNN':
        nbrs = sklearn.neighbors.NearestNeighbors(n_neighbors=cutoff+1).fit(coor)
        distances, indices = nbrs.kneighbors(coor)
        KNN_list = []
        for it in range(indices.shape[0]):
            KNN_list.append(pd.DataFrame(zip([it]*indices.shape[1],indices[it,:], distances[it,:])))
    
    if use_method == 'Radius':
        nbrs = sklearn.neighbors.NearestNeighbors(radius=cutoff).fit(coor)
        distances, indices = nbrs.radius_neighbors(coor, return_distance=True)
        KNN_list = []
        for it in range(indices.shape[0]):
            KNN_list.append(pd.DataFrame(zip([it]*indices[it].shape[0], indices[it], distances[it])))
            
    KNN_df = pd.concat(KNN_list)
    KNN_df.columns = ['Cell_i', 'Cell_j', 'Distance']

    Spatial_Graph = KNN_df.copy()
    Spatial_Graph = Spatial_Graph.loc[Spatial_Graph['Distance']>0,]
    id_cell_trans = dict(zip(range(coor.shape[0]), np.array(coor.index), ))
    Spatial_Graph['Cell_i'] = Spatial_Graph['Cell_i'].map(id_cell_trans)
    Spatial_Graph['Cell_j'] = Spatial_Graph['Cell_j'].map(id_cell_trans)
    print('The graph contains %d edges, %d cells.' %(Spatial_Graph.shape[0], adata.n_obs))
    print('%.4f neighbors per cell on average.' %(Spatial_Graph.shape[0]/adata.n_obs))

    adata.uns['Spatial_Graph'] = Spatial_Graph
    
    if use_method == 'Radius':
        Stats_Spatial_Graph(adata)



def Mutil_Construct_Spatial_Graph(adata, use_method='KNN', cutoff=None, obs_key=None, use_list=None):
    """
    Construct spatial graphs for either all or a subset of values in adata.obs[obs_key].

    Parameters:
    adata : AnnData
        The input AnnData object containing spatial coordinates in the obsm['spatial'] attribute.

    use_method : str (default='KNN')
        The method used to construct the spatial graph.
        Acceptable values: 'KNN' or 'Radius'.

    cutoff : int
        For 'KNN' method, it determines the number of nearest neighbors.
        For 'Radius' method, it determines the distance threshold for considering neighbors.

    obs_key : str
        The key in adata.obs used to access unique values.

    use_list : list or None (default=None)
        A list of specific values in adata.obs[obs_key] to be processed.
        If None, all unique values in adata.obs[obs_key] will be processed.

    Returns:
    AnnData
        A modified AnnData object with a 'Spatial_Graph' uns attribute containing the combined spatial graphs.
    """
    
    if use_list is None:
        # Process all unique values in adata.obs[obs_key]
        all_values = adata.obs[obs_key].unique()
        all_nets = []  # Store individual Spatial_Graph DataFrames
        for value in all_values:
            #temp_adata = adata.loc[adata.obs[obs_key] == value, :].copy()
            temp_adata = adata[adata.obs[obs_key] == value].copy()
            Construct_Spatial_Graph(temp_adata, use_method=use_method, cutoff=cutoff)
            all_nets.append(temp_adata.uns['Spatial_Graph'])
        
        # Combine all Spatial_Graph DataFrames
        combined_net = pd.concat(all_nets)
        adata.uns['Spatial_Graph'] = combined_net

    else:
        # Process a subset of values defined by use_list
        selected_values = set(use_list)
        selected_nets = []  # Store individual Spatial_Graph DataFrames
        for value in selected_values:
            temp_adata = adata[adata.obs[obs_key] == value].copy()
            Construct_Spatial_Graph(temp_adata, use_method=use_method, cutoff=cutoff)
            selected_nets.append(temp_adata.uns['Spatial_Graph'])
        
        # Combine selected Spatial_Graph DataFrames
        combined_net = pd.concat(selected_nets)
        adata.uns['Spatial_Graph'] = combined_net
        
        # Only keep rows in adata where adata.obs[obs_key] is in use_list
        adata = adata[adata.obs[obs_key].isin(use_list)].copy()
    
    return adata

def concat_adatas(adatas, section_list):
    """
    Process and store a list of AnnData objects in a dictionary.

    Parameters:
        adatas (list): List of AnnData objects.
        section_list (list): List of section IDs.

    Returns:
        dict: Dictionary containing processed AnnData objects.
    """
    
    if len(adatas) != len(section_list):
        raise ValueError("Length of adatas and section_list should be the same.")

    adata_list = {}

    for idx, (adata, section_id) in enumerate(zip(adatas, section_list)):
        # Add slice_id to obs using section_id
        adata.obs['slice_id'] = section_id

        # Make the spot name unique
        adata.obs_names = [f"{x}_{section_id}" for x in adata.obs_names]

        # Store the processed adata in the dictionary
        adata_list[section_id] = adata.copy()

    return adata_list

def merge_adatas(adatas, slice_keys):
    """
    Merge multiple AnnData objects, capturing and merging 'var' and 'uns' information present in any given object.
    
    Parameters:
        adatas (list): List of AnnData objects
        slice_keys (list): List of keys to mark each AnnData object

    Returns:
        AnnData: Merged AnnData object
    """
    if len(adatas) != len(slice_keys):
        raise ValueError("Length of adatas and section_list should be the same.")
    
    # Add 'slice_id' to the obs of each AnnData object using the provided slice_keys
    for adata, slice_key in zip(adatas, slice_keys):
        adata.obs['slice_id'] = slice_key
    
    # Use the concatenate function to merge all AnnData objects
    merged_adata = sc.concat(adatas)
    """
    # Merge 'var' attributes
    for adata in adatas:
        if hasattr(adata, 'var') and not adata.var.empty:
            for key, value in adata.var.items():
                if key not in merged_adata.var:
                    merged_adata.var[key] = value
                    
    # Merge 'uns' attributes
    for adata in adatas:
        if hasattr(adata, 'uns') and adata.uns:
            for key, value in adata.uns.items():
                if key not in merged_adata.uns:
                    merged_adata.uns[key] = value
    """
    return merged_adata


def clean_adata(adata_raw, save_obs=[]):
    """
    Clean the input AnnData object by retaining only specific attributes.
    
    Parameters:
        adata_raw (AnnData): The original AnnData object to be cleaned.
        save_obs (list, optional): A list of keys for the obs attributes to retain.
        
    Returns:
        AnnData: Cleaned AnnData object.
    """
    
    # Extract required attributes
    X = adata_raw.X
    var_names = adata_raw.var_names
    obs_names = adata_raw.obs_names
    spatial = adata_raw.obsm['spatial']
    
    # Create a new AnnData object
    adata_cleaned = ad.AnnData(X)
    adata_cleaned.var_names = var_names
    adata_cleaned.obs_names = obs_names
    adata_cleaned.obsm['spatial'] = spatial
    
    # Retain the obs attributes specified by save_obs
    for key in save_obs:
        if key in adata_raw.obs:
            adata_cleaned.obs[key] = adata_raw.obs[key]
    
    # Preserve 'spatial' key in uns if it exists
    if 'spatial' in adata_raw.uns:
        adata_cleaned.uns['spatial'] = adata_raw.uns['spatial']
        
    return adata_cleaned


def Stats_Spatial_Graph(adata):
    import matplotlib.pyplot as plt
    import pandas as pd
    
    # Existing calculations
    Num_edge = adata.uns['Spatial_Graph']['Cell_i'].shape[0]
    Mean_edge = Num_edge/adata.shape[0]
    plot_df = pd.value_counts(pd.value_counts(adata.uns['Spatial_Graph']['Cell_i']))
    plot_df = plot_df/adata.shape[0]
    
    # Setting up the figure and axis
    fig, ax = plt.subplots(figsize=[5, 4]) # Increase figure size for better readability
    bars = ax.bar(plot_df.index, plot_df, color='skyblue') # Using a softer color
    
    # Adding data values on top of the bars
    for bar in bars:
        yval = bar.get_height()
        ax.text(bar.get_x() + bar.get_width()/2, yval + 0.005, round(yval, 2), ha='center', va='bottom')
    
    # Enhancing axis labels and title
    plt.ylabel('Percentage', fontsize=12, fontweight='bold')
    plt.xlabel('Number of Neighbors', fontsize=12, fontweight='bold')
    plt.title('Number of Neighbors (Mean=%.2f)' % Mean_edge, fontsize=14, fontweight='bold')
    
    # Adjusting aesthetics
    plt.grid(axis='y', linestyle='--', alpha=0.7)
    ax.spines['top'].set_visible(False) # Hide top border
    ax.spines['right'].set_visible(False) # Hide right border
    
    plt.tight_layout()
    plt.show()
"""
def Stats_Spatial_Graph(adata):
    import matplotlib.pyplot as plt
    Num_edge = adata.uns['Spatial_Graph']['Cell_i'].shape[0]
    Mean_edge = Num_edge/adata.shape[0]
    plot_df = pd.value_counts(pd.value_counts(adata.uns['Spatial_Graph']['Cell_i']))
    plot_df = plot_df/adata.shape[0]
    fig, ax = plt.subplots(figsize=[3,2])
    plt.ylabel('Percentage')
    plt.xlabel('')
    plt.title('Number of Neighbors (Mean=%.2f)'%Mean_edge)
    ax.bar(plot_df.index, plot_df)
    
"""

def mclust_R(adata, num_cluster, add_key = 'mclust', modelNames='EEE', used_obsm='BINARY', random_seed=2020):
    """
    Cluster an AnnData object using the mclust algorithm from R.
    
    Parameters:
    ----------
    adata : AnnData
        The AnnData object containing the data to be clustered.
        
    num_cluster : int
        The number of clusters to be constructed.
        
    add_key : str, optional (default='mclust')
        The key added to `adata.obs` to store the cluster results.
        
    modelNames : str, optional (default='EEE')
        The parameter for the Mclust method in R specifying the model to be used for clustering. 
        For details on the possible models, refer to the mclust documentation.
        
    used_obsm : str, optional (default='BINARY')
        The key in `adata.obsm` that refers to the data matrix used for clustering.
        
    random_seed : int, optional (default=2020)
        The random seed for reproducibility.
        
    Returns:
    -------
    AnnData
        The input AnnData object with added clustering results in `adata.obs[add_key]`.
    """
    
    np.random.seed(random_seed)
    import rpy2.robjects as robjects
    robjects.r.library("mclust")

    import rpy2.robjects.numpy2ri
    rpy2.robjects.numpy2ri.activate()
    r_random_seed = robjects.r['set.seed']
    r_random_seed(random_seed)
    rmclust = robjects.r['Mclust']

    res = rmclust(rpy2.robjects.numpy2ri.numpy2rpy(adata.obsm[used_obsm]), num_cluster, modelNames)
    mclust_res = np.array(res[-2])

    adata.obs[add_key] = mclust_res
    adata.obs[add_key] = adata.obs[add_key].astype('int')
    adata.obs[add_key] = adata.obs[add_key].astype('category')
    return adata



"""
Utilities for the chord diagram.
"""

from collections import defaultdict
import numpy as np


def dist(points):
    '''
    Compute the distance between two points.

    Parameters
    ----------
    points : array of length 4
        The coordinates of the two points, P1 = (x1, y1) and P2 = (x2, y2)
        in the order [x1, y1, x2, y2].
    '''
    x1, y1 = points[0]
    x2, y2 = points[1]

    return np.sqrt((x2 - x1)*(x2 - x1) + (y2 - y1)*(y2 - y1))


def polar2xy(r, theta):
    '''
    Convert the coordinates of a point P from polar (r, theta) to cartesian
    (x, y).
    '''
    return np.array([r*np.cos(theta), r*np.sin(theta)])


def compute_positions(mat, deg, in_deg, out_deg, start_at, is_sparse, sort,
                      directed, extent, pad, arc, rotation, nodePos, pos):
    '''
    Compute all arcs and chords start/end positions.

    Parameters
    ----------
    mat : matrix
        Original matrix.
    deg : array
        Out (if undirected) or total (if directed) degree to compute the
        starting positions.
    in_deg : array
        In-degree to compute the end positions (if directed).
    out_deg : array
        Out-degree to compute the end positions (if directed).
    y : array
        Used to compute the arcs' ends.
    start_at : float
        Start of the first arc.
    is_sparse : bool
        Whether the matrix is sparse.
    sort : bool
        Sorting method.
    directed : bool
        Whether the chords are directed.
    extent : float in ]0, 360]
        Angular aperture of the diagram.
    pad : float
        Gap between entries.
    arc : list
        Used to store the arcs start and endpoints.
    rotation : list
        Store the rotation booleans for the names.
    nodePos : list
        Store the name positions.
    pos : dict
        Store the start and end positions for each arc under the form:
        (start1, end1, start2, end2), where (start1, end1) are the limits of the
        chords starting point, and (start2, end2) are the limits of the chord's
        end point.
    '''
    num_nodes = len(deg)

    # find position for each start and end
    y = deg / np.sum(deg).astype(float) * (extent - pad * num_nodes)

    if directed:
        y_out = out_deg / np.sum(deg).astype(float) * (extent - pad * num_nodes)

    starts = [start_at] + (
        start_at + np.cumsum(y + pad*np.ones(num_nodes))).tolist()

    out_ends = [s + d for s, d in zip(starts, (y_out if directed else y))]


    # relative positions within an arc
    zmat = [
        _get_normed_line(mat, i, out_deg if directed else deg, starts[i],
                         out_ends[i], is_sparse)
        for i in range(num_nodes)
    ]

    zin_mat = zmat

    if directed:
        zin_mat = [
            _get_normed_line(mat.T, i, in_deg, out_ends[i], starts[i + 1] - pad,
                             is_sparse)
            for i in range(num_nodes)
        ]

    # sort
    mat_ids = _get_sorted_ids(sort, zmat, num_nodes, directed)

    # compute positions
    for i in range(num_nodes):
        # arcs
        start = starts[i]
        end = start + y[i]
        arc.append((start, end))
        angle = 0.5*(start + end)

        if -30 <= (angle % 360) <= 180:
            angle -= 90
            rotation.append(False)
        else:
            angle -= 270
            rotation.append(True)

        nodePos.append(
            tuple(polar2xy(1.05, 0.5*(start + end)*np.pi/180.)) + (angle,))

        # sort chords
        z = zmat[i]
        z0 = start

        for j in mat_ids[i]:
            # compute the arrival points
            zj = zin_mat[j]
            startj = out_ends[j] if directed else starts[j]

            jids = mat_ids[j]

            distsort = (sort == "distance")

            if directed and not distsort:
                jids = jids[::-1]

            stop = np.where(np.equal(jids, i))[0][0]

            startji = startj + zj[jids[:stop]].sum()

            if distsort and directed:
                # we want j first for target positions
                startji += zj[j]

            if distsort and directed and i == j:
                pos[(i, j)] = (z0, z0 + z[j], startj, startj + zj[j])
            else:
                pos[(i, j)] = (z0, z0 + z[j], startji, startji + zj[jids[stop]])

            z0 += z[j]


# In-file functions

def _get_normed_line(mat, i, x, start, end, is_sparse):
    if is_sparse:
        row = mat.getrow(i).todense().A1
        return (row / x[i]) * (end - start)

    return (mat[i, :] / x[i]) * (end - start)


def _get_sorted_ids(sort, zmat, num_nodes, directed):
    mat_ids = defaultdict(lambda: list(range(num_nodes)))

    if sort == "size":
        mat_ids = [np.argsort(z) for z in zmat]
    elif sort == "distance":
        mat_ids = []
        for i in range(num_nodes):
            remainder = 0 if num_nodes % 2 else -1

            ids  = list(range(i - int(0.5*num_nodes), i))[::-1]

            if not directed:
                ids += [i]

            ids += list(range(i + int(0.5*num_nodes) + remainder, i, -1))

            if directed:
                ids += [i]

            # put them back into [0, num_nodes[
            ids = np.array(ids)
            ids[ids < 0] += num_nodes
            ids[ids >= num_nodes] -= num_nodes

            mat_ids.append(ids)
    elif sort is not None:
        raise ValueError("Invalid `sort`: '{}'".format(sort))

    return mat_ids


import numpy as np
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.font_manager as fm
import matplotlib.patches as mpatches
import random
import scanpy as sc
import networkx as nx
import pandas as pd
import anndata
from sklearn.cluster import KMeans
from scipy.spatial import ConvexHull
import seaborn as sns
from datetime import datetime,timedelta
import pkg_resources
from pkg_resources import DistributionNotFound, VersionConflict
import tomli
import os


from datetime import datetime, timedelta
import warnings
import platform
import os
try:
    import torch  # Optional, used for GPU information
except ImportError:  # pragma: no cover - optional dependency
    torch = None

sc_color=[
 '#1F577B', '#A56BA7', '#E0A7C8', '#E069A6', '#941456', '#FCBC10', '#EF7B77', '#279AD7','#F0EEF0',
 '#EAEFC5', '#7CBB5F','#368650','#A499CC','#5E4D9A','#78C2ED','#866017', '#9F987F','#E0DFED',
 '#01A0A7', '#75C8CC', '#F0D7BC', '#D5B26C', '#D5DA48', '#B6B812', '#9DC3C3', '#A89C92', '#FEE00C', '#FEF2A1']

red_color=['#F0C3C3','#E07370','#CB3E35','#A22E2A','#5A1713','#D3396D','#DBC3DC','#85539B','#5C2B80','#5C4694']
green_color=['#91C79D','#8FC155','#56AB56','#2D5C33','#BBCD91','#6E944A','#A5C953','#3B4A25','#010000']
orange_color=['#EFBD49','#D48F3E','#AC8A3E','#7D7237','#745228','#E1C085','#CEBC49','#EBE3A1','#6C6331','#8C9A48','#D7DE61']
blue_color=['#347862','#6BBBA0','#81C0DD','#3E8CB1','#88C8D2','#52B3AD','#265B58','#B2B0D4','#5860A7','#312C6C']
purple_color=['#823d86','#825b94','#bb98c6','#c69bc6','#a69ac9','#c5a6cc','#caadc4','#d1c3d4']

#more beautiful colors
# 28-color palettes with distinct neighboring colors
palette_28 = [
    '#E63946', '#1D3557', '#FFB703', '#2A9D8F', '#9B2226', '#4361EE', '#FF9F1C', '#560BAD',
    '#06D6A0', '#BC4749', '#4895EF', '#F4D03F', '#7209B7', '#52B788', '#D00000', '#4CC9F0',
    '#F7B801', '#3C096C', '#40916C', '#DC2F02', '#48CAE4', '#F9C74F', '#240046', '#081C15',
    '#9D0208', '#90E0EF', '#F3722C', '#14213D'
]

# 56-color palette with clear transitions
palette_56 = [
    '#001219', '#94D2BD', '#AE2012', '#3A0CA3', '#FF7B00', '#006466', '#E63946', '#1D3557',
    '#FFB703', '#2A9D8F', '#9B2226', '#4361EE', '#FF9F1C', '#560BAD', '#06D6A0', '#BC4749',
    '#4895EF', '#F4D03F', '#7209B7', '#52B788', '#D00000', '#4CC9F0', '#F7B801', '#3C096C',
    '#40916C', '#DC2F02', '#48CAE4', '#F9C74F', '#240046', '#081C15', '#9D0208', '#90E0EF',
    '#2D00F7', '#E76F51', '#006400', '#FF4D6D', '#073B4C', '#FF9E00', '#440154', '#55A630',
    '#7B2CBF', '#FF4800', '#0077B6', '#F72585', '#3D405B', '#588157', '#6A040F', '#023E8A',
    '#FF006E', '#2B9348', '#8338EC', '#F94144', '#0F4C5C', '#E85D04', '#540B0E', '#1A759F'
]

# 112-color palette with distinct transitions
palette_112 = [
    '#001219', '#94D2BD', '#AE2012', '#3A0CA3', '#FF7B00', '#006466', '#E63946', '#1D3557',
    '#FFB703', '#2A9D8F', '#9B2226', '#4361EE', '#FF9F1C', '#560BAD', '#06D6A0', '#BC4749',
    '#4895EF', '#F4D03F', '#7209B7', '#52B788', '#D00000', '#4CC9F0', '#F7B801', '#3C096C',
    '#40916C', '#DC2F02', '#48CAE4', '#F9C74F', '#240046', '#081C15', '#9D0208', '#90E0EF',
    '#2D00F7', '#E76F51', '#006400', '#FF4D6D', '#073B4C', '#FF9E00', '#440154', '#55A630',
    '#7B2CBF', '#FF4800', '#0077B6', '#F72585', '#3D405B', '#588157', '#6A040F', '#023E8A',
    '#FF006E', '#2B9348', '#8338EC', '#F94144', '#0F4C5C', '#E85D04', '#540B0E', '#1A759F',
    '#FF0A54', '#2B9348', '#5E60CE', '#F8961E', '#073B3A', '#FF6B6B', '#2D3047', '#70E000',
    '#5A189A', '#FF7900', '#0096C7', '#FF48B0', '#344E41', '#606C38', '#641220', '#03045E',
    '#FF0075', '#386641', '#7400B8', '#F3722C', '#27474E', '#F48C06', '#3C1518', '#0077B6',
    '#FF758F', '#40916C', '#6930C3', '#F9844A', '#1B4965', '#FAA307', '#582F0E', '#0096C7',
    '#FF477E', '#2D6A4F', '#5E60CE', '#F9C74F', '#004E89', '#FF9500', '#50514F', '#70E000',
    '#9D4EDD', '#FF6D00', '#219EBC', '#FF5C8A', '#344E41', '#606C38', '#800E13', '#0353A4',
    '#FF0A54', '#387D44', '#480CA8', '#F3622C', '#1A5653', '#FF9F1C', '#4A4E69', '#55A630'
]

# Vibrant palette with clear distinctions (24 colors)
vibrant_palette = [
    '#FF0000', '#00FF00', '#0000FF', '#FFFF00', '#FF00FF', '#00FFFF', 
    '#FF8000', '#FF0080', '#80FF00', '#00FF80', '#8000FF', '#0080FF',
    '#FF3333', '#33FF33', '#3333FF', '#FFFF33', '#FF33FF', '#33FFFF',
    '#FF9933', '#FF3399', '#99FF33', '#33FF99', '#9933FF', '#3399FF'
]

# Earth tones palette (24 colors)
earth_palette = [
    '#8B4513', '#DAA520', '#556B2F', '#2F4F4F', '#8B008B', '#4682B4',
    '#CD853F', '#BDB76B', '#6B8E23', '#4F666A', '#800080', '#4169E1',
    '#D2691E', '#F0E68C', '#9ACD32', '#5F9EA0', '#9932CC', '#1E90FF',
    '#A0522D', '#EEE8AA', '#698B22', '#008B8B', '#9400D3', '#00BFFF'
]

# Pastel palette (24 colors)
pastel_palette = [
    '#FFB3BA', '#BAFFC9', '#BAE1FF', '#FFFFBA', '#FFDFBA', '#E0BBE4',
    '#957DAD', '#D291BC', '#FEC8D8', '#FFDFD3', '#B5EAD7', '#C7CEEA',
    '#FFB997', '#F5B0CB', '#D4F0F0', '#FFF5BA', '#A8E6CF', '#DBB4D8',
    '#FFD3B5', '#D4F0F0', '#B5EAD7', '#E2F0CB', '#C7CEEA', '#FFDAC1'
]






sc_color_cmap = LinearSegmentedColormap.from_list('Custom', sc_color, len(sc_color))



omics="""
   ____            _     _    __                  
  / __ \____ ___  (_)___| |  / /__  _____________ 
 / / / / __ `__ \/ / ___/ | / / _ \/ ___/ ___/ _ \ 
/ /_/ / / / / / / / /__ | |/ /  __/ /  (__  )  __/ 
\____/_/ /_/ /_/_/\___/ |___/\___/_/  /____/\___/                                              
"""
days_christmas="""
      .
   __/ \__
   \     /
   /.'o'.\
    .o.'.         Merry Christmas!
   .'.'o'.
  o'.o.'.o.
 .'.o.'.'.o.       ____ 
.o.'.o.'.o.'.     / __ \____ ___  (_)___| |  / /__  _____________ 
   [_____]       / / / / __ `__ \/ / ___/ | / / _ \/ ___/ ___/ _ \ 
    \___/       / /_/ / / / / / / / /__ | |/ /  __/ /  (__  )  __/ 
                \____/_/ /_/ /_/_/\___/ |___/\___/_/  /____/\___/
"""
#Tua Xiong
days_chinese_new_year="""
                                        ,   ,
                                        $,  $,     ,
                                        "ss.$ss. .s'
                                ,     .ss$$$$$$$$$$s,
                                $. s$$$$$$$$$$$$$$`$$Ss
                                "$$$$$$$$$$$$$$$$$$o$$$       ,
                               s$$$$$$$$$$$$$$$$$$$$$$$$s,  ,s
                              s$$$$$$$$$"$$$$$$ssss$$$$$$"$$$$$,
                              s$$$$$$$$$$sss$$$$ssssss"$$$$$$$$"
                             s$$$$$$$$$$'         `\"\"\"ss"$"$s\"\"
                             s$$$$$$$$$$,              `\"\"\"\"\"$  .s$$s
                             s$$$$$$$$$$$$s,...               `s$$'  `
                         `ssss$$$$$$$$$$$$$$$$$$$$####s.     .$$"$.   , s-
                           `""\""$$$$$$$$$$$$$$$$$$$$#####$$$$$$"     $.$'
                                 "$$$$$$$$$$$$$$$$$$$$$####s""     .$$$|
                                  "$$$$$$$$$$$$$$$$$$$$$$$$##s    .$$" $
                                   $$""$$$$$$$$$$$$$$$$$$$$$$$$$$$$$"   `
                                  $$"  "$"$$$$$$$$$$$$$$$$$$$$S""\""'
                             ,   ,"     '  $$$$$$$$$$$$$$$$####s
                             $.          .s$$$$$$$$$$$$$$$$$####"
                 ,           "$s.   ..ssS$$$$$$$$$$$$$$$$$$$####"
                 $           .$$$S$$$$$$$$$$$$$$$$$$$$$$$$#####"
                 Ss     ..sS$$$$$$$$$$$$$$$$$$$$$$$$$$$######""
                  "$$sS$$$$$$$$$$$$$$$$$$$$$$$$$$$########"
           ,      s$$$$$$$$$$$$$$$$$$$$$$$$#########""'
           $    s$$$$$$$$$$$$$$$$$$$$$#######""'      s'         ,
           $$..$$$$$$$$$$$$$$$$$$######"'       ....,$$....    ,$
            "$$$$$$$$$$$$$$$######"' ,     .sS$$$$$$$$$$$$$$$$s$$
              $$$$$$$$$$$$#####"     $, .s$$$$$$$$$$$$$$$$$$$$$$$$s.
   )          $$$$$$$$$$$#####'      `$$$$$$$$$###########$$$$$$$$$$$.
  ((          $$$$$$$$$$$#####       $$$$$$$$###"       "####$$$$$$$$$$
  ) \         $$$$$$$$$$$$####.     $$$$$$###"             "###$$$$$$$$$   s'
 (   )        $$$$$$$$$$$$$####.   $$$$$###"                ####$$$$$$$$s$$'
 )  ( (       $$"$$$$$$$$$$$#####.$$$$$###' -OmicVerse     .###$$$$$$$$$$"
 (  )  )   _,$"   $$$$$$$$$$$$######.$$##'                .###$$$$$$$$$$
 ) (  ( \.         "$$$$$$$$$$$$$#######,,,.          ..####$$$$$$$$$$$"
(   )$ )  )        ,$$$$$$$$$$$$$$$$$$####################$$$$$$$$$$$"
(   ($$  ( \     _sS"  `"$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$S$$,
 )  )$$$s ) )  .      .   `$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$"'  `$$
  (   $$$Ss/  .$,    .$,,s$$$$$$##S$$$$$$$$$$$$$$$$$$$$$$$$S""        '
    \)_$$$$$$$$$$$$$$$$$$$$$$$##"  $$        `$$.        `$$.
        `"S$$$$$$$$$$$$$$$$$#"      $          `$          `$
            `\"""\""\""\""\""\""'         '           '           '
"""

spring_festival = { 
    2022: datetime(2022, 2, 1), 
    2023: datetime(2023, 1, 22), 
    2024: datetime(2024, 2, 10), # ... 
    2025: datetime(2025, 1, 29),
    2026: datetime(2026, 2, 17),
    2027: datetime(2027, 2, 6),
}

try:
    from importlib.metadata import version
except ModuleNotFoundError:
    from pkg_resources import get_distribution
    version = lambda name: get_distribution(name).version

name = "omicverse"
try:
    __version__ = version(name)
except Exception:
    __version__ = "unknown"

_has_printed_logo = False  # Flag to ensure logo prints only once
#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# emoji map for status reporting
EMOJI = {
    "start":        "🔬",  # experiment start
    "deps":         "🔗",  # dependency check
    "settings":     "⚙️",  # configure settings
    "warnings":     "🚫",  # suppress warnings
    "gpu":          "🧬",  # GPU check
    "logo":         "🌟",  # print logo
    "done":         "✅",  # done
}


def plot_set(verbosity: int = 3, dpi: int = 80, 
             facecolor: str = 'white', font_path: str = None):
    r"""Configure plotting settings for OmicVerse.
    
    Sets up scanpy verbosity, matplotlib parameters, suppresses warnings,
    detects GPU devices, and displays the OmicVerse logo.
    
    Arguments:
        verbosity: Scanpy verbosity level (3)
        dpi: Resolution for matplotlib figures (80) 
        facecolor: Background color for figures ('white')
        font_path: Path to font for custom fonts (None)
        
    Returns:
        None
    """
    global _has_printed_logo

    print(f"{EMOJI['start']} Starting plot initialization...")

    # 1) dependency check
    #print(f"{EMOJI['deps']} Checking dependencies...")
    #check_dependencies()
    # print(f"{EMOJI['done']} Dependencies OK")

    # 2) scanpy verbosity & figure params
    #print(f"{EMOJI['settings']} Applying plotting settings (verbosity={verbosity}, dpi={dpi})")
    sc.settings.verbosity = verbosity
    sc.settings.set_figure_params(dpi=dpi, facecolor=facecolor)
    #print(f"{EMOJI['done']} Settings applied")

    # 3) Custom font setup
    if font_path is not None:
        # Check if user wants Arial font (auto-download)
        if font_path.lower() in ['arial', 'arial.ttf'] and not font_path.endswith('.ttf'):
            try:
                # Create a persistent cache location for the Arial font
                import tempfile
                import requests

                cache_dir = tempfile.gettempdir()
                cached_arial_path = os.path.join(cache_dir, 'omicverse_arial.ttf')
                
                # Check if Arial font is already cached
                if os.path.exists(cached_arial_path):
                    print(f"Using already downloaded Arial font from: {cached_arial_path}")
                    font_path = cached_arial_path
                else:
                    print("Downloading Arial font from GitHub...")
                    arial_url = "https://github.com/kavin808/arial.ttf/raw/refs/heads/master/arial.ttf"
                    
                    # Download the font
                    response = requests.get(arial_url, timeout=30)
                    response.raise_for_status()
                    
                    # Save the font to cache location
                    with open(cached_arial_path, 'wb') as f:
                        f.write(response.content)
                    
                    # Use the cached font file
                    font_path = cached_arial_path
                    print(f"Arial font downloaded successfully to: {cached_arial_path}")
                
            except Exception as e:
                print(f"Failed to download Arial font: {e}")
                print("Continuing with default font settings...")
                font_path = None
        
        if font_path is not None:
            try:
                # 1) Create a brand-new manager
                fm.fontManager = fm.FontManager()
                
                # 2) Add your file
                fm.fontManager.addfont(font_path)
                
                # 3) Now find out what name it uses
                name = fm.FontProperties(fname=font_path).get_name()
                print("Registered as:", name)
                
                # 4) Point rcParams at that name
                mpl.rcParams['font.family'] = 'sans-serif'
                mpl.rcParams['font.sans-serif'] = [name, 'DejaVu Sans']
                
            except Exception as e:
                print(f"Failed to set custom font: {e}")
                print("Continuing with default font settings...")

    # 4) suppress user/future/deprecation warnings
    #print(f"{EMOJI['warnings']} Suppressing common warnings")
    warnings.simplefilter("ignore", category=UserWarning)
    warnings.simplefilter("ignore", category=FutureWarning)
    warnings.simplefilter("ignore", category=DeprecationWarning)
    #print(f"{EMOJI['done']} Warnings suppressed")

    # 5) GPU detection
    print(f"{EMOJI['gpu']} Detecting CUDA devices…")
    if torch is None or not torch.cuda.is_available():
        print(f"{EMOJI['warnings']} No CUDA devices found")
    else:
        try:
            for idx in range(torch.cuda.device_count()):
                props = torch.cuda.get_device_properties(idx)
                print(f"{EMOJI['done']} [GPU {idx}] {props.name}")
                print(f"    • Total memory: {props.total_memory/1024**3:.1f} GB")
                print(f"    • Compute capability: {props.major}.{props.minor}")
        except Exception as e:
            print(f"{EMOJI['warnings']} GPU detection failed: {e}")

    # 6) print logo & version only once
    if not _has_printed_logo:
        #print(f"{EMOJI['logo']} OmicVerse Logo:")
        today = datetime.now()
        chinese_new_year = spring_festival.get(today.year)
        if today.month == 12 and today.day in (24, 25):
            print(days_christmas)
        elif chinese_new_year and (chinese_new_year - timedelta(days=1) <= today <= chinese_new_year + timedelta(days=3)):
            print(days_chinese_new_year)
        else:
            print(omics)
        print(f"🔖 Version: {__version__}   📚 Tutorials: https://omicverse.readthedocs.io/")
        _has_printed_logo = True

    print(f"{EMOJI['done']} plot_set complete.\n")


# Create aliases for backward compatibility
plotset = plot_set
ov_plot_set = plot_set

def pyomic_palette()->list:
    r"""Returns the default OmicVerse color palette.
    
    Returns:
        List of hex color codes for plotting
    """ 
    return sc_color

def palette()->list:
    r"""Returns the default OmicVerse color palette.
    
    Returns:
        List of hex color codes for plotting
    """ 
    return sc_color

def red_palette()->list:
    r"""Returns a red-themed color palette.
    
    Returns:
        List of red-themed hex color codes
    """ 
    return red_color

def green_palette()->list:
    r"""Returns a green-themed color palette.
    
    Returns:
        List of green-themed hex color codes
    """ 
    return green_color

def orange_palette()->list:
    r"""Returns an orange-themed color palette.
    
    Returns:
        List of orange-themed hex color codes
    """ 
    return orange_color

def blue_palette()->list:
    r"""Returns a blue-themed color palette.
    
    Returns:
        List of blue-themed hex color codes
    """ 
    return blue_color

def plot_text_set(text, text_knock=2, text_maxsize=20):
    r"""Format text for plotting by adding line breaks.
    
    Arguments:
        text: Text string to format
        text_knock: Number of words between line breaks (2)
        text_maxsize: Maximum text length before formatting (20)
        
    Returns:
        Formatted text string with line breaks
    """
    if len(text) <= text_maxsize:
        return text
    
    words = text.split(' ')
    formatted_text = []
    for i, word in enumerate(words):
        if i > 0 and i % text_knock == 0:
            formatted_text.append('\n')
        formatted_text.append(word)
    
    return ' '.join(formatted_text).strip()

    
def ticks_range(x,width):
    r"""Generate tick positions for multi-group plots.
    
    Arguments:
        x: Number of ticks
        width: Width spacing between ticks
        
    Returns:
        List of tick positions
    """
    nticks=[]
    pticks=[]
    start=-(x//2)
    end=(x//2)
    for i in range(x//2):
        nticks.append(start+width)
        start+=width
        pticks.append(end-width)
        end-=width
    if x%2==0:
        ticks=nticks+pticks
    elif x%2==1:
        ticks=nticks+[0]+pticks
    return ticks

def plot_boxplot(data,hue,x_value,y_value,width=0.6,title='',
                 figsize=(6,3),palette=None,fontsize=10,
                 legend_bbox=(1, 0.55),legend_ncol=1,):
    r"""Create boxplot with jittered points for grouped data.
    
    Arguments:
        data: DataFrame containing the data to plot
        hue: Column name for grouping variable
        x_value: Column name for x-axis categories
        y_value: Column name for y-axis values
        width: Width of boxplots (0.6)
        title: Plot title ('')
        figsize: Figure size ((6,3))
        palette: List of colors (None)
        fontsize: Font size for labels (10)
        legend_bbox: Legend bounding box ((1, 0.55))
        legend_ncol: Number of legend columns (1)
        
    Returns:
        Tuple of (figure, axes) objects
    """

    #获取需要分割的数据
    hue=hue
    hue_datas=list(set(data[hue]))

    #获取箱线图的横坐标
    x=x_value
    ticks=list(set(data[x]))

    #在这个数据中，我们有6个不同的癌症，每个癌症都有2个基因（2个箱子）
    #所以我们需要得到每一个基因的6个箱线图位置，6个散点图的抖动
    plot_data1={}#字典里的每一个元素就是每一个基因的所有值
    plot_data_random1={}#字典里的每一个元素就是每一个基因的随机20个值
    plot_data_xs1={}#字典里的每一个元素就是每一个基因的20个抖动值


    #箱子的参数
    #width=0.6
    y=y_value
    for hue_data,num in zip(hue_datas,ticks_range(len(hue_datas),width)):
        data_a=[]
        data_a_random=[]
        data_a_xs=[]
        for i,k in zip(ticks,range(len(ticks))):
            test_data=data.loc[((data[x]==i)&(data[hue]==hue_data)),y].tolist()
            data_a.append(test_data)
            if len(test_data)<20:
                data_size=len(test_data)
            else:
                data_size=20
            random_data=random.sample(test_data,data_size)
            data_a_random.append(random_data)
            data_a_xs.append(np.random.normal(k*len(hue_datas)+num, 0.04, len(random_data)))
        #data_a=np.array(data_a)
        data_a_random=np.array(data_a_random)
        plot_data1[hue_data]=data_a 
        plot_data_random1[hue_data]=data_a_random
        plot_data_xs1[hue_data]=data_a_xs

    fig, ax = plt.subplots(figsize=figsize)
    #色卡
    if palette==None:
        palette=pyomic_palette()
    #palette=["#a64d79","#674ea7"]
    #绘制箱线图
    for hue_data,hue_color,num in zip(hue_datas,palette,ticks_range(len(hue_datas),width)):
        b1=ax.boxplot(plot_data1[hue_data], 
                    positions=np.array(range(len(ticks)))*len(hue_datas)+num, 
                    sym='', 
                    widths=width,)
        plt.setp(b1['boxes'], color=hue_color)
        plt.setp(b1['whiskers'], color=hue_color)
        plt.setp(b1['caps'], color=hue_color)
        plt.setp(b1['medians'], color=hue_color)

        clevels = np.linspace(0., 1., len(plot_data_random1[hue_data]))
        for x, val, clevel in zip(plot_data_xs1[hue_data], plot_data_random1[hue_data], clevels):
            plt.scatter(x, val,c=hue_color,alpha=0.4)

    #坐标轴字体
    #fontsize=10
    #修改横坐标
    ax.set_xticks(range(0, len(ticks) * len(hue_datas), len(hue_datas)), ticks,fontsize=fontsize)
    #修改纵坐标
    yticks=ax.get_yticks()
    ax.set_yticks(yticks[yticks>=0],yticks[yticks>=0],fontsize=fontsize)

    labels = hue_datas  #legend标签列表，上面的color即是颜色列表
    color = palette
    #用label和color列表生成mpatches.Patch对象，它将作为句柄来生成legend
    patches = [ mpatches.Patch(color=color[i], label="{:s}".format(labels[i]) ) for i in range(len(hue_datas)) ] 
    ax.legend(handles=patches,bbox_to_anchor=legend_bbox, ncol=legend_ncol,fontsize=fontsize)

    #设置标题
    ax.set_title(title,fontsize=fontsize+1)
    #设置spines可视化情况
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)
    return fig,ax

def plot_network(G:nx.Graph,G_type_dict:dict,G_color_dict:dict,pos_type:str='spring',pos_dim:int=2,
                figsize:tuple=(4,4),pos_scale:int=10,pos_k=None,pos_alpha:float=0.4,
                node_size:int=50,node_alpha:float=0.6,node_linewidths:int=1,
                plot_node=None,plot_node_num:int=20,
                label_verticalalignment:str='center_baseline',label_fontsize:int=12,
                label_fontfamily:str='Arial',label_fontweight:str='bold',label_bbox=None,
                legend_bbox:tuple=(0.7, 0.05),legend_ncol:int=3,legend_fontsize:int=12,
                legend_fontweight:str='bold'):
    r"""Plot network graph with customizable node and edge properties.
    
    Arguments:
        G: NetworkX graph object
        G_type_dict: Dictionary mapping nodes to types
        G_color_dict: Dictionary mapping nodes to colors
        pos_type: Layout algorithm - 'spring' or 'kamada_kawai' ('spring')
        pos_dim: Layout dimension - 2 or 3 (2)
        figsize: Figure size ((4,4))
        pos_scale: Layout scale factor (10)
        pos_k: Spring layout k parameter (None)
        pos_alpha: Edge transparency (0.4)
        node_size: Base node size (50)
        node_alpha: Node transparency (0.6)
        node_linewidths: Node border width (1)
        plot_node: Specific nodes to label (None)
        plot_node_num: Number of top degree nodes to label (20)
        label_verticalalignment: Label vertical alignment ('center_baseline')
        label_fontsize: Label font size (12)
        label_fontfamily: Label font family ('Arial')
        label_fontweight: Label font weight ('bold')
        label_bbox: Label bounding box properties (None)
        legend_bbox: Legend position ((0.7, 0.05))
        legend_ncol: Legend columns (3)
        legend_fontsize: Legend font size (12)
        legend_fontweight: Legend font weight ('bold')
        
    Returns:
        Tuple of (figure, axes) objects
    """
    

    fig, ax = plt.subplots(figsize=figsize)
    if pos_type=='spring':
        pos = nx.spring_layout(G, scale=pos_scale, k=pos_k)
    elif pos_type=='kamada_kawai':
        pos=nx.kamada_kawai_layout(G,dim=pos_dim,scale=pos_scale)
    degree_dict = dict(G.degree(G.nodes()))
    
    G_color_dict=dict(zip(G.nodes,[G_color_dict[i] for i in G.nodes]))
    G_type_dict=dict(zip(G.nodes,[G_type_dict[i] for i in G.nodes]))

    nx.draw_networkx_edges(G, pos,nodelist=list(G_color_dict.keys()), alpha=pos_alpha)
    nx.draw_networkx_nodes(
        G,
        pos,
        nodelist=list(G_color_dict.keys()),
        node_size=[degree_dict[v]*node_size for v in G],
        node_color=list(G_color_dict.values()),
        alpha=node_alpha,
        linewidths=node_linewidths,
    )
    if plot_node!=None:
        hub_gene=plot_node
    else:
        hub_gene=[i[0] for i in sorted(degree_dict.items(),key=lambda x: x[1],reverse=True)[:plot_node_num]]
    
    pos1=dict()
    #for i in pos.keys():
    #    pos1[i]=np.array([-1000,-1000])
    for i in hub_gene:
        pos1[i]=pos[i]
    #label_options = {"ec": "white", "fc": "white", "alpha": 0.6}
    #nx.draw_networkx_labels(
    #    G,pos1,verticalalignment=label_verticalalignment,
    #    font_size=label_fontsize,font_family=label_fontfamily,
    #    font_weight=label_fontweight,bbox=label_bbox,
    #)
    from adjustText import adjust_text
    import adjustText
    texts=[ax.text(pos1[i][0], 
               pos1[i][1],
               i,
               fontdict={'size':label_fontsize,'weight':label_fontweight,'color':'black'}
               ) for i in hub_gene if 'ENSG' not in i]
    if adjustText.__version__<='0.8':
        adjust_text(texts,only_move={'text': 'xy'},arrowprops=dict(arrowstyle='->', color='red'),)
    else:
        adjust_text(texts,only_move={"text": "xy", "static": "xy", "explode": "xy", "pull": "xy"},
                    arrowprops=dict(arrowstyle='->', color='red'))
   #adjust_text(texts,only_move={'text': 'xy'},arrowprops=dict(arrowstyle='->', color='red'),)

    ax.axis("off")
    
    t=pd.DataFrame(index=G_type_dict.keys())
    t['gene_type_dict']=G_type_dict
    t['gene_color_dict']=G_color_dict
    type_color_dict={}
    for i in t['gene_type_dict'].value_counts().index:
        type_color_dict[i]=t.loc[t['gene_type_dict']==i,'gene_color_dict'].values[0]
    
    patches = [ mpatches.Patch(color=type_color_dict[i], label="{:s}".format(i) ) for i in type_color_dict.keys() ] 

    plt.legend(handles=patches,bbox_to_anchor=legend_bbox, ncol=legend_ncol,fontsize=legend_fontsize)
    leg = plt.gca().get_legend() #或leg=ax.get_legend()
    ltext = leg.get_texts()
    plt.setp(ltext, fontsize=legend_fontsize,fontweight=legend_fontweight)
    
    return fig,ax

def plot_cellproportion(adata:anndata.AnnData,celltype_clusters:str,visual_clusters:str,
                       visual_li=None,visual_name:str='',figsize:tuple=(4,6),
                       ticks_fontsize:int=12,labels_fontsize:int=12,
                       legend:bool=False):
    r"""Plot stacked bar chart showing cell type proportions across groups.
    
    Arguments:
        adata: AnnData object
        celltype_clusters: Column name for cell types
        visual_clusters: Column name for grouping variable
        visual_li: List of groups to plot (None)
        visual_name: Label for x-axis ('')
        figsize: Figure size ((4,6))
        ticks_fontsize: Font size for tick labels (12)
        labels_fontsize: Font size for axis labels (12)
        legend: Whether to show legend (False)
        
    Returns:
        Tuple of (figure, axes) objects
    """

    b=pd.DataFrame(columns=['cell_type','value','Week'])
    
    if visual_li==None:
        adata.obs[visual_clusters]=adata.obs[visual_clusters].astype('category')
        visual_li=adata.obs[visual_clusters].cat.categories
    
    for i in visual_li:
        b1=pd.DataFrame()
        test=adata.obs.loc[adata.obs[visual_clusters]==i,celltype_clusters].value_counts()
        b1['cell_type']=test.index
        b1['value']=test.values/test.sum()
        b1['Week']=i.replace('Retinoblastoma_','')
        b=pd.concat([b,b1])
    
    plt_data2=adata.obs[celltype_clusters].value_counts()
    plot_data2_color_dict=dict(zip(adata.obs[celltype_clusters].cat.categories,adata.uns['{}_colors'.format(celltype_clusters)]))
    plt_data3=adata.obs[visual_clusters].value_counts()
    plot_data3_color_dict=dict(zip([i.replace('Retinoblastoma_','') for i in adata.obs[visual_clusters].cat.categories],adata.uns['{}_colors'.format(visual_clusters)]))
    b['cell_type_color'] = b['cell_type'].map(plot_data2_color_dict)
    b['stage_color']=b['Week'].map(plot_data3_color_dict)
    
    fig, ax = plt.subplots(figsize=figsize)
    #用ax控制图片
    #sns.set_theme(style="whitegrid")
    #sns.set_theme(style="ticks")
    n=0
    all_celltype=adata.obs[celltype_clusters].cat.categories
    for i in all_celltype:
        if n==0:
            test1=b[b['cell_type']==i]
            ax.bar(x=test1['Week'],height=test1['value'],width=0.8,color=list(set(test1['cell_type_color']))[0], label=i)
            bottoms=test1['value'].values
        else:
            test2=b[b['cell_type']==i]
            ax.bar(x=test2['Week'],height=test2['value'],bottom=bottoms,width=0.8,color=list(set(test2['cell_type_color']))[0], label=i)
            test1=test2
            bottoms+=test1['value'].values
        n+=1
    if legend!=False:
        plt.legend(bbox_to_anchor=(1.05, -0.05), loc=3, borderaxespad=0,fontsize=10)
    
    plt.grid(False)
    
    plt.grid(False)
    ax.spines['top'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(True)
    ax.spines['left'].set_visible(True)

    # 设置左边和下边的坐标刻度为透明色
    #ax.yaxis.tick_left()
    #ax.xaxis.tick_bottom()
    #ax.xaxis.set_tick_params(color='none')
    #ax.yaxis.set_tick_params(color='none')

    # 设置左边和下边的坐标轴线为独立的线段
    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))

    plt.xticks(fontsize=ticks_fontsize,rotation=90)
    plt.yticks(fontsize=ticks_fontsize)
    plt.xlabel(visual_name,fontsize=labels_fontsize)
    plt.ylabel('Cells per Stage',fontsize=labels_fontsize)
    fig.tight_layout()
    return fig,ax

def plot_embedding_celltype(adata:anndata.AnnData,figsize:tuple=(6,4),basis:str='umap',
                            celltype_key:str='major_celltype',title:str=None,
                            celltype_range:tuple=(2,9),
                            embedding_range:tuple=(3,10),
                            xlim:int=-1000)->tuple:
    r"""Create combined embedding plot with cell type legend and counts.
    
    Arguments:
        adata: AnnData object
        figsize: Figure size ((6,4))
        basis: Embedding basis name ('umap')
        celltype_key: Column name for cell types ('major_celltype')
        title: Plot title (None)
        celltype_range: Grid range for cell type panel ((2,9))
        embedding_range: Grid range for embedding panel ((3,10))
        xlim: X-axis limit for counts (-1000)
        
    Returns:
        Tuple of (figure, [embedding_axis, celltype_axis])
    """

    adata.obs[celltype_key]=adata.obs[celltype_key].astype('category')
    cell_num_pd=pd.DataFrame(adata.obs[celltype_key].value_counts())
    if '{}_colors'.format(celltype_key) in adata.uns.keys():
        cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories.tolist(),
                        adata.uns['{}_colors'.format(celltype_key)]))
    else:
        if len(adata.obs[celltype_key].cat.categories)>28:
            cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.default_102))
        else:
            cell_color_dict=dict(zip(adata.obs[celltype_key].cat.categories,sc.pl.palettes.zeileis_28))

    if figsize==None:
        if len(adata.obs[celltype_key].cat.categories)<10:
            fig = plt.figure(figsize=(6,4))
        else:
            print('The number of cell types is too large, please set the figsize parameter')
            return
    else:
        fig = plt.figure(figsize=figsize)
    grid = plt.GridSpec(10, 10)
    ax1 = fig.add_subplot(grid[:, embedding_range[0]:embedding_range[1]])       # 占据第一行的所有列
    ax2 = fig.add_subplot(grid[celltype_range[0]:celltype_range[1], :2]) 
    # 定义子图的大小和位置
         # 占据第二行的前两列
    #ax3 = fig.add_subplot(grid[1:, 2])      # 占据第二行及以后的最后一列
    #ax4 = fig.add_subplot(grid[2, 0])       # 占据最后一行的第一列
    #ax5 = fig.add_subplot(grid[2, 1])       # 占据最后一行的第二列

    sc.pl.embedding(
        adata,
        basis=basis,
        color=[celltype_key],
        title='',
        frameon=False,
        #wspace=0.65,
        ncols=3,
        ax=ax1,
        legend_loc=False,
        show=False
    )

    for idx,cell in zip(range(cell_num_pd.shape[0]),
                        adata.obs[celltype_key].cat.categories):
        ax2.scatter(100,
                cell,c=cell_color_dict[cell],s=50)
        ax2.plot((100,cell_num_pd.loc[cell,celltype_key]),(idx,idx),
                c=cell_color_dict[cell],lw=4)
        ax2.text(100,idx+0.2,
                cell+'('+str("{:,}".format(cell_num_pd.loc[cell,celltype_key]))+')',fontsize=11)
    ax2.set_xlim(xlim,cell_num_pd.iloc[1].values[0]) 
    ax2.text(xlim,idx+1,title,fontsize=12)
    ax2.grid(False)
    ax2.spines['top'].set_visible(False)
    ax2.spines['right'].set_visible(False)
    ax2.spines['bottom'].set_visible(False)
    ax2.spines['left'].set_visible(False)
    ax2.axis('off')

    return fig,[ax1,ax2]

def gen_mpl_labels(
    adata, groupby, exclude=(), 
    basis='X_umap',ax=None, adjust_kwargs=None, text_kwargs=None
):
    r"""Add cluster labels at median positions in embedding plots.
    
    Arguments:
        adata: AnnData object
        groupby: Column name for grouping
        exclude: Groups to exclude from labeling (())
        basis: Embedding basis name ('X_umap')
        ax: Matplotlib axes object (None)
        adjust_kwargs: Parameters for text adjustment (None)
        text_kwargs: Parameters for text styling (None)
        
    Returns:
        None
    """
    if adjust_kwargs is None:
        adjust_kwargs = {"text_from_points": False}
    if text_kwargs is None:
        text_kwargs = {}

    medians = {}

    for g, g_idx in adata.obs.groupby(groupby).groups.items():
        if g in exclude:
            continue
        medians[g] = np.median(adata[g_idx].obsm[basis], axis=0)

    if ax is None:
        texts = [
            plt.text(x=x, y=y, s=k, **text_kwargs) for k, (x, y) in medians.items()
        ]
    else:
        texts = [ax.text(x=x, y=y, s=k, **text_kwargs) for k, (x, y) in medians.items()]
    from adjustText import adjust_text
    adjust_text(texts, **adjust_kwargs)

def plot_embedding(adata:anndata.AnnData,basis:str,color:str,color_dict=None,
                   figsize:tuple=(4,4),**kwargs):
    r"""Create embedding plot with customizable colors.
    
    Arguments:
        adata: AnnData object
        basis: Embedding basis name
        color: Column name for coloring
        color_dict: Custom color mapping (None)
        figsize: Figure size ((4,4))
        **kwargs: Additional parameters for sc.pl.embedding
        
    Returns:
        Tuple of (figure, axes) objects
    """
    if type(color)!=str:
        print("Only one color could be input, don't input list")
        return
    fig,ax=plt.subplots(1,1,figsize=figsize)
    adata.obs[color]=adata.obs[color].astype('category')

    if '{}_colors'.format(color) in adata.uns.keys():
        print('{}_colors'.format(color))
        type_color_all=dict(zip(adata.obs[color].cat.categories,adata.uns['{}_colors'.format(color)]))
    else:
        if len(adata.obs[color].cat.categories)>28:
            type_color_all=dict(zip(adata.obs[color].cat.categories,sc.pl.palettes.default_102))
        else:
            type_color_all=dict(zip(adata.obs[color].cat.categories,sc.pl.palettes.zeileis_28))
    if color_dict is not None:
        for color_key in color_dict.keys():
            type_color_all[color_key]=color_dict[color_key]
    
    adata.uns['{}_colors'.format(color)]=np.array([i for i in type_color_all.values()])
    sc.pl.embedding(adata,basis=basis,
                    color=color,ax=ax,**kwargs)
    return fig,ax

from sklearn.preprocessing import MinMaxScaler

def normalize_to_minus_one_to_one(arr):
    # 将数组reshape为二维数组，因为MinMaxScaler接受二维数据
    arr = arr.reshape(-1, 1)
    
    # 创建MinMaxScaler对象，并设定归一化的范围为[-1, 1]
    scaler = MinMaxScaler(feature_range=(-1, 1))
    
    # 对数据进行归一化处理
    normalized_arr = scaler.fit_transform(arr)
    
    # 将结果转换回一维数组
    normalized_arr = normalized_arr.flatten()
    
    return normalized_arr

def stacking_vol(data_dict:dict,color_dict:dict,
                 pval_threshold:float=0.01,
                 log2fc_threshold:int=2,
                 figsize:tuple=(8,4),
                 sig_color:str='#a51616',
                 normal_color:str='#c7c7c7',
                 plot_genes_num:int=10,
                 plot_genes_fontsize:int=8,
                plot_genes_weight:str='bold')->tuple:
    r"""Create stacked volcano plots for comparing multiple datasets.
    
    Arguments:
        data_dict: Dictionary with DataFrames containing 'logfoldchanges', 'pvals_adj', 'names'
        color_dict: Dictionary mapping dataset names to colors
        pval_threshold: P-value significance threshold (0.01)
        log2fc_threshold: Log2 fold change threshold (2)
        figsize: Figure size ((8,4))
        sig_color: Color for significant points ('#a51616')
        normal_color: Color for non-significant points ('#c7c7c7')
        plot_genes_num: Number of top genes to label (10)
        plot_genes_fontsize: Font size for gene labels (8)
        plot_genes_weight: Font weight for gene labels ('bold')
        
    Returns:
        Tuple of (figure, axes_dict)
    """
    
    fig = plt.figure(figsize=figsize)
    grid = plt.GridSpec(len(data_dict.keys())*2, len(data_dict.keys())*2)
    axes={}
    j_before=0
    y_min,y_max=0,0
    for i in data_dict.keys():
        y_min=min(y_min,data_dict[i]['logfoldchanges'].min())
        y_max=max(y_max,data_dict[i]['logfoldchanges'].max())

    for i,j in zip(data_dict.keys(),
               range(2,len(data_dict.keys())*2+2,2)):
        print(j_before,j)
        axes[i]=fig.add_subplot(grid[:, j_before:j])
        j_before+=2
    
        x=np.random.normal(0, 1, data_dict[i].shape[0])
        x=normalize_to_minus_one_to_one(x)

        plot_data=pd.DataFrame()
        plot_data['logfoldchanges']=data_dict[i]['logfoldchanges']
        plot_data['pvals_adj']=data_dict[i]['pvals_adj']
        plot_data['abslogfoldchanges']=abs(data_dict[i]['logfoldchanges'])
        plot_data['sig']='normal'
        plot_data.loc[(plot_data['pvals_adj']<pval_threshold)&(plot_data['abslogfoldchanges']>log2fc_threshold),'sig']='sig'
        plot_data['x']=x
        plot_data.index=data_dict[i]['names']


        axes[i].scatter(plot_data.loc[plot_data['sig']!='sig','x'],
                   plot_data.loc[plot_data['sig']!='sig','logfoldchanges'],
                   color=normal_color,alpha=0.5)

        axes[i].scatter(plot_data.loc[plot_data['sig']=='sig','x'],
                   plot_data.loc[plot_data['sig']=='sig','logfoldchanges'],
                   color=sig_color,alpha=0.8)

        axes[i].axhspan(0-log2fc_threshold/2, log2fc_threshold/2, 
                        facecolor=color_dict[i], alpha=1)

        
        axes[i].set_ylim(y_min,y_max)

        plt.grid(False)
        plt.yticks(fontsize=12)

        axes[i].spines['top'].set_visible(False)
        if j_before!=2:
            axes[i].spines['left'].set_visible(False)
            axes[i].axis('off')
        axes[i].spines['bottom'].set_visible(False)
        axes[i].spines['right'].set_visible(False)
        axes[i].set_ylim(y_min,y_max)
        axes[i].set_xticks([])
        
        hub_gene=plot_data.loc[plot_data['sig']=='sig'].sort_values('abslogfoldchanges',
                                                                    ascending=False).index[:plot_genes_num]
        from adjustText import adjust_text
        texts=[axes[i].text(plot_data.loc[gene,'x'], 
                            plot_data.loc[gene,'logfoldchanges'],
                            gene,
                            fontdict={'size':plot_genes_fontsize,
                                    'weight':plot_genes_weight,
                                     'color':'black'}) 
               for gene in hub_gene]
        adjust_text(texts,only_move={'text': 'xy'},arrowprops=dict(arrowstyle='->', color='red'),)
            
    return fig,axes


def plot_ConvexHull(adata:anndata.AnnData,basis:str,cluster_key:str,
                    hull_cluster:str,ax,color=None,alpha:float=0.2):
    r"""Add convex hull outline for a specific cluster in embedding plot.
    
    Arguments:
        adata: AnnData object
        basis: Embedding basis name in adata.obsm
        cluster_key: Column name for cluster assignments
        hull_cluster: Specific cluster to outline
        ax: Matplotlib axes object
        color: Hull color (None for automatic)
        alpha: Hull transparency (0.2)
        
    Returns:
        Modified matplotlib axes object
    """
    
    adata.obs[cluster_key]=adata.obs[cluster_key].astype('category')
    if '{}_colors'.format(cluster_key) in adata.uns.keys():
        print('{}_colors'.format(cluster_key))
        type_color_all=dict(zip(adata.obs[cluster_key].cat.categories,adata.uns['{}_colors'.format(cluster_key)]))
    else:
        if len(adata.obs[cluster_key].cat.categories)>28:
            type_color_all=dict(zip(adata.obs[cluster_key].cat.categories,sc.pl.palettes.default_102))
        else:
            type_color_all=dict(zip(adata.obs[cluster_key].cat.categories,sc.pl.palettes.zeileis_28))
    
    #color_dict=dict(zip(adata.obs[cluster_key].cat.categories,adata.uns[f'{cluster_key}_colors']))
    points=adata[adata.obs[cluster_key]==hull_cluster].obsm[basis]
    hull = ConvexHull(points)
    vert = np.append(hull.vertices, hull.vertices[0])  # close the polygon by appending the first point at the end
    if color==None:
        ax.plot(points[vert, 0], points[vert, 1], '--', c=type_color_all[hull_cluster])
        ax.fill(points[vert, 0], points[vert, 1], c=type_color_all[hull_cluster], alpha=alpha)
    else:
        ax.plot(points[vert, 0], points[vert, 1], '--', c=color)
        ax.fill(points[vert, 0], points[vert, 1], c=color, alpha=alpha)
    return ax





class geneset_wordcloud(object):

    def __init__(self,adata,cluster_key,pseudotime,resolution=1000,figsize=(4,10)):
        self.adata=adata
        self.cluster_key=cluster_key
        self.pseudotime=pseudotime
        self.figsize=figsize
        self.resolution=resolution

    def get(self,):
        from wordcloud import WordCloud, STOPWORDS, ImageColorGenerator
        #Get the DataFrame of anndata
        test_df=self.adata.to_df()
        
        #Calculate the mean of each cluster
        ct_pd=pd.DataFrame(columns=test_df.columns)
        for ct in self.adata.obs[self.cluster_key].unique():
            ct_pd.loc[ct]=test_df.loc[self.adata.obs.loc[self.adata.obs[self.cluster_key]==ct].index].mean(axis=0)
    
        # 遍历每个基因，找到最高表达的细胞类型
        max_expr_cell_types = []
        for gene in ct_pd.columns:
            max_expr_cell_type = ct_pd[gene].idxmax()
            max_expr_cell_types.append((gene, max_expr_cell_type))
        
        # 将结果转换为数据框
        result_df = pd.DataFrame(max_expr_cell_types, columns=['Gene', 'Max_Expression_Cell_Type'])
    
        
        size_dict=dict(result_df['Max_Expression_Cell_Type'].value_counts()/result_df.shape[0])

        self.adata.obs[self.cluster_key]=self.adata.obs[self.cluster_key].astype('category')
        if '{}_colors'.format(self.cluster_key) in self.adata.uns.keys():
            cell_color_dict=dict(zip(self.adata.obs[self.cluster_key].cat.categories.tolist(),
                            self.adata.uns['{}_colors'.format(self.cluster_key)]))
        else:
            if len(self.adata.obs[self.cluster_key].cat.categories)>28:
                cell_color_dict=dict(zip(self.adata.obs[self.cluster_key].cat.categories,sc.pl.palettes.default_102))
            else:
                cell_color_dict=dict(zip(self.adata.obs[self.cluster_key].cat.categories,sc.pl.palettes.zeileis_28))


        wc_dict={}
        for ct in self.adata.obs[self.cluster_key].unique():
            #print(ct)
            word_li=result_df.loc[result_df['Max_Expression_Cell_Type']==ct,'Gene'].values.tolist()
            print(ct,100*self.figsize[0],
                  int(100*size_dict[ct]*self.figsize[1]))
            wc = WordCloud(background_color="#FFFFFF",min_font_size=12,max_font_size=700, max_words=30,
                           width=100*self.figsize[0],
                           height=int(100*size_dict[ct]*self.figsize[1]),
                           contour_width=3, contour_color='firebrick')
            # 生成词云
            wc.generate(''.join([i.split(' (')[0] for i in word_li]))
            wc_dict[ct]=wc

        self.wc_dict=wc_dict.copy()
        self.size_dict=size_dict
        self.result_df=result_df
        self.color_dict=cell_color_dict
        return wc_dict

    def get_geneset(self):
        return self.result_df

    def get_wordcloud(self):
        return self.wc_dict

    def plot(self):
        fig = plt.figure(figsize=self.figsize)
        grid = plt.GridSpec(self.resolution, 10)
        
        import matplotlib.colors as mcolors
        
        last_idx=0
        
        for idx,ct in zip(range(len(self.adata.obs[self.cluster_key].unique())),
                          self.adata.obs.groupby(self.cluster_key)[self.pseudotime].mean().sort_values().index):
            next_idx=last_idx+self.size_dict[ct]
            print(ct,round(last_idx*self.resolution),round(next_idx*self.resolution))
            ax=fig.add_subplot(grid[round(last_idx*self.resolution):round(next_idx*self.resolution), :])      # 占据第二行的前两列
        
            colors=['#FFFFFF',self.color_dict[ct]]
            xcmap = mcolors.LinearSegmentedColormap.from_list('test_cmap', colors, N=100)
            
            ax.imshow(self.wc_dict[ct].recolor(colormap=xcmap), interpolation='bilinear')
            last_idx+=self.size_dict[ct]
            #ax.grid(False)
            if idx!=0:
                ax.axhline(y=0, c="#000000")
            ax.axis(False)
            # 绘制边框
            ax.spines['top'].set_visible(True)
            ax.spines['bottom'].set_visible(True)
            ax.spines['left'].set_visible(True)
            ax.spines['right'].set_visible(True)

        return fig
    def plot_heatmap(self,n_convolve=10,figwidth=10,cmap='RdBu_r',
                     cbar=False,cbar_kws=None,cbar_fontsize=12):
        if cbar_kws==None:
            cbar_kws={'shrink':0.5,'location':'left'}

        fig = plt.figure(figsize=(figwidth,self.figsize[1]))
        grid = plt.GridSpec(self.resolution, 10)
        
        import matplotlib.colors as mcolors
        
        last_idx=0
        
        for idx,ct in zip(range(len(self.adata.obs[self.cluster_key].unique())),
                          self.adata.obs.groupby(self.cluster_key)[self.pseudotime].mean().sort_values().index):
            next_idx=last_idx+self.size_dict[ct]
            #print(ct,round(last_idx*self.resolution),round(next_idx*self.resolution))
            ax=fig.add_subplot(grid[round(last_idx*self.resolution):round(next_idx*self.resolution), figwidth-self.figsize[0]:])      # 占据第二行的前两列
        
            colors=['#FFFFFF',self.color_dict[ct]]
            xcmap = mcolors.LinearSegmentedColormap.from_list('test_cmap', colors, N=100)
            
            ax.imshow(self.wc_dict[ct].recolor(colormap=xcmap), interpolation='bilinear')
            last_idx+=self.size_dict[ct]
            #ax.grid(False)
            if idx!=0:
                ax.axhline(y=0, c="#000000")
            ax.axis(False)
            # 绘制边框
            ax.spines['top'].set_visible(True)
            ax.spines['bottom'].set_visible(True)
            ax.spines['left'].set_visible(True)
            ax.spines['right'].set_visible(True)


        #sort time
        time = self.adata.obs[self.pseudotime].values
        time = time[np.isfinite(time)]
        
        from scipy.sparse import issparse
        X = self.adata.X
        if issparse(X):
            X = X.A
        df = pd.DataFrame(X[np.argsort(time)], columns=self.adata.var_names)

        #convolve
        
        if n_convolve is not None:
            weights = np.ones(n_convolve) / n_convolve
            for gene in self.adata.var_names:
                # TODO: Handle exception properly
                try:
                    df[gene] = np.convolve(df[gene].values, weights, mode="same")
                except ValueError as e:
                    print(f"Skipping variable {gene}: {e}")
        max_sort = np.argsort(np.argmax(df.values, axis=0))
        df = pd.DataFrame(df.values[:, max_sort], columns=df.columns[max_sort])

        from sklearn.preprocessing import MinMaxScaler
        
        scaler = MinMaxScaler()
        normalized_data = scaler.fit_transform(df)
        
        ax2=fig.add_subplot(grid[:, :figwidth-self.figsize[0]])
        sns.heatmap(normalized_data.T,cmap=cmap,ax=ax2,cbar=cbar,cbar_kws=cbar_kws)
        # use matplotlib.colorbar.Colorbar object
        if cbar!=False:
            cbar1 = ax2.collections[0].colorbar
            # here set the labelsize by 20
            cbar1.ax.tick_params(labelsize=cbar_fontsize)
        #ax2.imshow(normalized_data.T,cmap='RdBu_r',)
        ax2.grid(False)
        ax2.axis(False)
        
        #ax3=fig.add_subplot(grid[:10, :8])
        # 添加类别可视化（以不同颜色的矩形表示）
        category_colors = self.adata.obs[self.cluster_key].map(self.color_dict).values[np.argsort(time)]
        for i, color in enumerate(category_colors):
            rect = plt.Rectangle((i, 0), 2, 2, color=color)
            ax2.add_patch(rect)

        return fig


from scanpy.plotting._anndata import ranking
from scanpy.plotting._utils import savefig_or_show
def plot_pca_variance_ratio(
    adata,
    use_rep='scaled|original|pca_var_ratios',
    n_pcs: int = 30,
    log: bool = False,
    show=None,
    save=None,
):
    ranking(
        adata,
        "uns",
        use_rep,
        n_points=n_pcs,
        #dictionary="pca",
        labels="PC",
        log=log,
    )
    savefig_or_show("pca_variance_ratio", show=show, save=save)

def plot_pca_variance_ratio1(adata,threshold=0.85):

    import matplotlib.pyplot as plt
    plt.scatter(range(len(adata.uns['scaled|original|pca_var_ratios'])),
                adata.uns['scaled|original|pca_var_ratios'])
    ratio_max=max(adata.uns['scaled|original|pca_var_ratios'])
    ratio_max_85=(1-threshold)*ratio_max
    pcs_85_num=len(adata.uns['scaled|original|pca_var_ratios'][adata.uns['scaled|original|pca_var_ratios']>ratio_max_85])
    plt.axhline(ratio_max_85)
    plt.title(f'PCs:{pcs_85_num}')
    plt.xlabel('ranking')


def check_dependencies(dependencies=None, check_full=False):
    r"""Check if installed package versions match requirements.
    
    Arguments:
        dependencies: List of dependency strings (None)
        check_full: Whether to check optional dependencies (False)
        
    Returns:
        None
    """
    if dependencies is None:
        try:
            # Get the directory of the current file
            current_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
            pyproject_path = os.path.join(current_dir, 'pyproject.toml')
            
            with open(pyproject_path, 'rb') as f:
                pyproject = tomli.load(f)
                dependencies = pyproject['project']['dependencies']

            # If check_full is True, also check full dependencies
            if check_full and 'project' in pyproject and 'optional-dependencies' in pyproject['project']:
                if 'full' in pyproject['project']['optional-dependencies']:
                    full_deps = pyproject['project']['optional-dependencies']['full']
                    dependencies.extend(full_deps)

        except Exception as e:
            print(f"Warning: Could not read dependencies from pyproject.toml: {e}")
            return

    try:
        pkg_resources.require(dependencies)
        print("All dependencies are satisfied.")
    except (DistributionNotFound, VersionConflict) as e:
        print(f"Dependency error: {e}")


import os
import pandas as pd
import numpy as np
from tqdm import trange
import scanpy as sc
import scipy.sparse as sp
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.pyplot import MultipleLocator
from .utils import var_stabilize
import math



def plot_gene(adata, platform="visium", sorted_by='PI', input_data=None, size=2, 
              top_n=25, ncols_each_sheet=5, nrows_each_sheet=5,save_path=None):
    if sp.issparse(adata.X):
        gene_data = adata.X.A.T
    else:
        gene_data = adata.X.T
    if platform=="visium":
        gene_data = var_stabilize(gene_data)

    x = adata.obsm["spatial"][:,1] 
    y = adata.obsm["spatial"][:,0]
    
    ncols = ncols_each_sheet
    nrows = nrows_each_sheet
    
    if platform=="visium" or platform=="ST":
        cmp = mpl.colors.LinearSegmentedColormap.from_list('pink_green', ["#3AB370","#EAE7CC","#FD1593"], N=256)
    else:
        nodes = [0.0, 0.05, 0.3, 1.0]
        cmp = mpl.colors.LinearSegmentedColormap.from_list("mycmap", list(zip(nodes, ["#EAE7CC","#EAE7CC","#FD1593","#FD1593"])))

    mpl.rcParams['mathtext.fontset'] = 'custom'
    mpl.rcParams['mathtext.rm'] = 'Arial:italic:bold'
    mpl.rcParams['mathtext.it'] = 'Arial:italic:bold'     
    
    if sorted_by in ["SEP", "SIG", "PI"]:
        pi = pd.DataFrame(adata.var[sorted_by].reset_index())
        # sort gene according to 
        sorted_list = pi.sort_values(by = sorted_by, ascending = False)
        sorted_list.columns = ["geneID",sorted_by]
        sorted_idx = sorted_list.index.values
        gene_data = gene_data[sorted_idx, :]        
        gene_data = gene_data[:top_n]
        sorted_genename = sorted_list["geneID"].reset_index(drop = True)
        sorted_pi = sorted_list[sorted_by].reset_index(drop = True)
    else:
        adata.var_names = [i.upper() for i in list(adata.var_names)]
        assert input_data["genename"] is not None or input_data["qvalue"] is not None,"please set \" input_data.columns=['genename','qvalue']\" "
        input_data = input_data.sort_values("qvalue")
        sorted_genename = input_data["genename"]
        sorted_genename = pd.Series([i.upper() for i in sorted_genename])
        sorted_idx = np.ones(sorted_genename.shape).astype(int)
        for j in range(sorted_genename.shape[0]):
            sorted_idx[j] = np.where(adata.var_names==sorted_genename[j])[0]       
        gene_data = gene_data[sorted_idx, :]
        gene_data = gene_data[:top_n]
        sorted_qval = input_data["qvalue"]
        sorted_qval = sorted_qval.reset_index(drop = True)
    print("\nDrawing pictures:")    
    for figs in trange(0, len(gene_data), ncols * nrows):
        if figs < len(gene_data):
            n = figs
        else:
            n = int(len(gene_data)/ ncols/ nrows) * ncols * nrows + 1
        figure, axs = plt.subplots(nrows, ncols, figsize=(12,16))
        
        for i in range(ncols):
            for j in range(nrows):
                label = gene_data[n, :]
                if sorted_by == None:
                    genename = adata.var_names[n]
                elif sorted_by in ["SEP", "SIG", "PI"]:
                    genename = sorted_genename[n]
                    pi_value = sorted_pi[n]
                else:
                    genename = sorted_genename[n]
                    qval = sorted_qval[n]
                # axs = axs.T
                ax = axs[i][j]

                ax.scatter(y, x, s = size, c = label, cmap = cmp)

                if sorted_by == None:   
                    ax.set_title('${}$'.format(genename))
                elif sorted_by in ["SEP", "SIG", "PI"]:
                    ax.set_title('${}$'.format(genename) + ' ({}={:.3f})'.format(sorted_by,pi_value), fontdict = {'family': 'Arial','weight': 'bold','style': 'normal','size': 11})
                else:
                    ax.set_title('${}$'.format(genename) + ' (q-value={:g})'.format(qval), fontdict = {'family': 'Arial','weight': 'bold','style': 'normal','size': 11})
                ax.set_xlim(min(y)-20, max(y)+20)
                ax.set_ylim(min(x)-20, max(x)+20)
            
                x_major_locator = MultipleLocator(2000)
                y_major_locator = MultipleLocator(2000)                
                ax.xaxis.set_major_locator(x_major_locator)
                ax.yaxis.set_major_locator(y_major_locator)
                if platform=="visium" or platform=="SeqFISH" or platform=="ST":
                    ax.invert_yaxis()               
                ax.set_aspect('equal', 'box')                
                n += 1
               
                if n >= len(gene_data):
                    break
            if n >= len(gene_data):
                    break
                
        for i in range(nrows):
            for j in range(ncols):
                ax = axs[i][j]
                ax.set_xticks([])
                ax.set_yticks([])
                ax.spines['top'].set_visible(False) 
                ax.spines['bottom'].set_visible(False) 
                ax.spines['left'].set_visible(False)  
                ax.spines['right'].set_visible(False)
                
        plt.axis('off')
        plt.tight_layout()
        plt.subplots_adjust()
        if save_path:
            if not os.path.isdir("{}/SVGs".format(save_path)):
                os.makedirs("{}/SVGs".format(save_path))
            plt.savefig("{}/SVGs/P{}.png".format(save_path, math.ceil(n/(ncols*nrows))), dpi=600, bbox_inches='tight')       
    print("\nDrawing completed !!")
    

    
    


# This file is a modified version of the scrublet.py file from scanpy.
# The only difference is that it allows for the use of GPU.
# The original file can be found at:
# https://github.com/scverse/scanpy/blob/main/scanpy/preprocessing/_scrublet/core.py
from __future__ import annotations

from importlib.util import find_spec
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from anndata import AnnData
from scipy import sparse

from scanpy import logging as logg
from scanpy import preprocessing as pp
from scanpy._compat import old_positionals
from scanpy.get import _get_obs_rep
from scanpy.preprocessing._scrublet import pipeline
from scanpy.preprocessing._scrublet.core import Scrublet
from .._settings import EMOJI


from importlib.util import find_spec
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from anndata import AnnData
from scipy import sparse

from scanpy import logging as logg
from scanpy import preprocessing as pp
from scanpy._compat import old_positionals
from scanpy.get import _get_obs_rep
from scanpy.preprocessing._scrublet import pipeline
from scanpy.preprocessing._scrublet.core import Scrublet

if TYPE_CHECKING:
    from scanpy._compat import _LegacyRandom
    from scanpy.neighbors import _Metric, _MetricFn


@old_positionals(
    "batch_key",
    "sim_doublet_ratio",
    "expected_doublet_rate",
    "stdev_doublet_rate",
    "synthetic_doublet_umi_subsampling",
    "knn_dist_metric",
    "normalize_variance",
    "log_transform",
    "mean_center",
    "n_prin_comps",
    "use_approx_neighbors",
    "get_doublet_neighbor_parents",
    "n_neighbors",
    "threshold",
    "verbose",
    "copy",
    "random_state",
)
def scrublet(
    adata: AnnData,
    adata_sim: AnnData | None = None,
    *,
    batch_key: str | None = None,
    sim_doublet_ratio: float = 2.0,
    expected_doublet_rate: float = 0.05,
    stdev_doublet_rate: float = 0.02,
    synthetic_doublet_umi_subsampling: float = 1.0,
    knn_dist_metric: _Metric | _MetricFn = "euclidean",
    normalize_variance: bool = True,
    log_transform: bool = False,
    mean_center: bool = True,
    n_prin_comps: int = 30,
    use_approx_neighbors: bool | None = None,
    get_doublet_neighbor_parents: bool = False,
    n_neighbors: int | None = None,
    threshold: float | None = None,
    verbose: bool = True,
    copy: bool = False,
    random_state: _LegacyRandom = 0,
    use_gpu: bool = False,
) -> AnnData | None:
    r"""Predict doublets using Scrublet with optional GPU acceleration.

    Predict cell doublets using a nearest-neighbor classifier of observed
    transcriptomes and simulated doublets. This implementation includes
    GPU acceleration options for improved performance on large datasets.

    Arguments:
        adata: Annotated data matrix of shape n_obs × n_vars
        adata_sim (AnnData): Pre-simulated doublets (default: None)
        batch_key (str): Column name for batch information (default: None)
        sim_doublet_ratio (float): Number of doublets to simulate relative to observed cells (default: 2.0)
        expected_doublet_rate (float): Estimated doublet rate for the experiment (default: 0.05)
        stdev_doublet_rate (float): Uncertainty in expected doublet rate (default: 0.02)
        synthetic_doublet_umi_subsampling (float): UMI sampling rate for synthetic doublets (default: 1.0)
        knn_dist_metric (str): Distance metric for nearest neighbors (default: 'euclidean')
        normalize_variance (bool): Whether to normalize gene variance (default: True)
        log_transform (bool): Whether to log-transform data prior to PCA (default: False)
        mean_center (bool): Whether to center data for PCA (default: True)
        n_prin_comps (int): Number of principal components for embedding (default: 30)
        use_approx_neighbors (bool): Use approximate nearest neighbor search (default: None)
        get_doublet_neighbor_parents (bool): Return parent transcriptomes for doublet neighbors (default: False)
        n_neighbors (int): Number of neighbors for KNN graph (default: None)
        threshold (float): Doublet score threshold for classification (default: None)
        verbose (bool): Whether to log progress updates (default: True)
        copy (bool): Return copy instead of modifying in place (default: False)
        random_state (int): Random seed for reproducibility (default: 0)
        use_gpu (bool): Whether to use GPU acceleration (default: False)

    Returns:
        adata with doublet predictions added if copy=False, otherwise returns modified copy
    """
    if threshold is None and not find_spec("skimage"):  # pragma: no cover
        # Scrublet.call_doublets requires `skimage` with `threshold=None` but PCA
        # is called early, which is wasteful if there is not `skimage`
        msg = "threshold is None and thus scrublet requires skimage, but skimage is not installed."
        raise ValueError(msg)

    if copy:
        adata = adata.copy()

    start = logg.info(f"Running Scrublet{EMOJI['start']}")

    adata_obs = adata.copy()

    def _run_scrublet(ad_obs: AnnData, ad_sim: AnnData | None = None):
        # With no adata_sim we assume the regular use case, starting with raw
        # counts and simulating doublets

        if ad_sim is None:
            pp.filter_genes(ad_obs, min_cells=3)
            pp.filter_cells(ad_obs, min_genes=3)

            # Doublet simulation will be based on the un-normalised counts, but on the
            # selection of genes following normalisation and variability filtering. So
            # we need to save the raw and subset at the same time.

            ad_obs.layers["raw"] = ad_obs.X.copy()
            pp.normalize_total(ad_obs)

            # HVG process needs log'd data.
            ad_obs.layers["log1p"] = ad_obs.X.copy()
            pp.log1p(ad_obs, layer="log1p")
            pp.highly_variable_genes(ad_obs, layer="log1p")
            del ad_obs.layers["log1p"]
            ad_obs = ad_obs[:, ad_obs.var["highly_variable"]].copy()

            # Simulate the doublets based on the raw expressions from the normalised
            # and filtered object.

            ad_sim = scrublet_simulate_doublets(
                ad_obs,
                layer="raw",
                sim_doublet_ratio=sim_doublet_ratio,
                synthetic_doublet_umi_subsampling=synthetic_doublet_umi_subsampling,
                random_seed=random_state,
            )
            del ad_obs.layers["raw"]
            if log_transform:
                pp.log1p(ad_obs)
                pp.log1p(ad_sim)

            # Now normalise simulated and observed in the same way

            pp.normalize_total(ad_obs, target_sum=1e6)
            pp.normalize_total(ad_sim, target_sum=1e6)

        ad_obs = _scrublet_call_doublets(
            adata_obs=ad_obs,
            adata_sim=ad_sim,
            n_neighbors=n_neighbors,
            expected_doublet_rate=expected_doublet_rate,
            stdev_doublet_rate=stdev_doublet_rate,
            mean_center=mean_center,
            normalize_variance=normalize_variance,
            n_prin_comps=n_prin_comps,
            use_approx_neighbors=use_approx_neighbors,
            knn_dist_metric=knn_dist_metric,
            get_doublet_neighbor_parents=get_doublet_neighbor_parents,
            threshold=threshold,
            random_state=random_state,
            verbose=verbose,
        )

        return {"obs": ad_obs.obs, "uns": ad_obs.uns["scrublet"]}

    if batch_key is not None:
        if batch_key not in adata.obs.columns:
            msg = (
                "`batch_key` must be a column of .obs in the input AnnData object,"
                f"but {batch_key!r} is not in {adata.obs.keys()!r}."
            )
            raise ValueError(msg)

        # Run Scrublet independently on batches and return just the
        # scrublet-relevant parts of the objects to add to the input object

        batches = np.unique(adata.obs[batch_key])
        scrubbed = [
            _run_scrublet(
                adata_obs[adata_obs.obs[batch_key] == batch].copy(),
                adata_sim,
            )
            for batch in batches
        ]
        scrubbed_obs = pd.concat([scrub["obs"] for scrub in scrubbed])

        # Now reset the obs to get the scrublet scores

        adata.obs = scrubbed_obs.loc[adata.obs_names.values]

        # Save the .uns from each batch separately

        adata.uns["scrublet"] = {}
        adata.uns["scrublet"]["batches"] = dict(
            zip(batches, [scrub["uns"] for scrub in scrubbed])
        )

        # Record that we've done batched analysis, so e.g. the plotting
        # function knows what to do.

        adata.uns["scrublet"]["batched_by"] = batch_key

    else:
        scrubbed = _run_scrublet(adata_obs, adata_sim)

        # Copy outcomes to input object from our processed version

        adata.obs["doublet_score"] = scrubbed["obs"]["doublet_score"]
        adata.obs["predicted_doublet"] = scrubbed["obs"]["predicted_doublet"]
        adata.uns["scrublet"] = scrubbed["uns"]

    logg.info(f"    Scrublet finished{EMOJI['done']}", time=start)

    return adata if copy else None


def _scrublet_call_doublets(
    adata_obs: AnnData,
    adata_sim: AnnData,
    *,
    n_neighbors: int | None = None,
    expected_doublet_rate: float = 0.05,
    stdev_doublet_rate: float = 0.02,
    mean_center: bool = True,
    normalize_variance: bool = True,
    n_prin_comps: int = 30,
    use_approx_neighbors: bool | None = None,
    knn_dist_metric: _Metric | _MetricFn = "euclidean",
    get_doublet_neighbor_parents: bool = False,
    threshold: float | None = None,
    random_state: _LegacyRandom = 0,
    verbose: bool = True,
) -> AnnData:
    """Core function for predicting doublets using Scrublet :cite:p:`Wolock2019`.

    Predict cell doublets using a nearest-neighbor classifier of observed
    transcriptomes and simulated doublets.

    Parameters
    ----------
    adata_obs
        The annotated data matrix of shape ``n_obs`` × ``n_vars``. Rows
        correspond to cells and columns to genes. Should be normalised with
        :func:`~scanpy.pp.normalize_total` and filtered to include only highly
        variable genes.
    adata_sim
        Anndata object generated by
        :func:`~scanpy.pp.scrublet_simulate_doublets`, with same number of vars
        as adata_obs. This should have been built from adata_obs after
        filtering genes and cells and selcting highly-variable genes.
    n_neighbors
        Number of neighbors used to construct the KNN graph of observed
        transcriptomes and simulated doublets. If ``None``, this is
        automatically set to ``np.round(0.5 * np.sqrt(n_obs))``.
    expected_doublet_rate
        The estimated doublet rate for the experiment.
    stdev_doublet_rate
        Uncertainty in the expected doublet rate.
    mean_center
        If True, center the data such that each gene has a mean of 0.
        `sklearn.decomposition.PCA` will be used for dimensionality
        reduction.
    normalize_variance
        If True, normalize the data such that each gene has a variance of 1.
        `sklearn.decomposition.TruncatedSVD` will be used for dimensionality
        reduction, unless `mean_center` is True.
    n_prin_comps
        Number of principal components used to embed the transcriptomes prior
        to k-nearest-neighbor graph construction.
    use_approx_neighbors
        Use approximate nearest neighbor method (annoy) for the KNN
        classifier.
    knn_dist_metric
        Distance metric used when finding nearest neighbors. For list of
        valid values, see the documentation for annoy (if `use_approx_neighbors`
        is True) or sklearn.neighbors.NearestNeighbors (if `use_approx_neighbors`
        is False).
    get_doublet_neighbor_parents
        If True, return the parent transcriptomes that generated the
        doublet neighbors of each observed transcriptome. This information can
        be used to infer the cell states that generated a given
        doublet state.
    threshold
        Doublet score threshold for calling a transcriptome a doublet. If
        `None`, this is set automatically by looking for the minimum between
        the two modes of the `doublet_scores_sim_` histogram. It is best
        practice to check the threshold visually using the
        `doublet_scores_sim_` histogram and/or based on co-localization of
        predicted doublets in a 2-D embedding.
    random_state
        Initial state for doublet simulation and nearest neighbors.
    verbose
        If :data:`True`, log progress updates.

    Returns
    -------
    if ``copy=True`` it returns or else adds fields to ``adata``:

    ``.obs['doublet_score']``
        Doublet scores for each observed transcriptome

    ``.obs['predicted_doublets']``
        Boolean indicating predicted doublet status

    ``.uns['scrublet']['doublet_scores_sim']``
        Doublet scores for each simulated doublet transcriptome

    ``.uns['scrublet']['doublet_parents']``
        Pairs of ``.obs_names`` used to generate each simulated doublet transcriptome

    ``.uns['scrublet']['parameters']``
        Dictionary of Scrublet parameters

    """
    # Estimate n_neighbors if not provided, and create scrublet object.

    if n_neighbors is None:
        n_neighbors = int(round(0.5 * np.sqrt(adata_obs.shape[0])))

    # Note: Scrublet() will sparse adata_obs.X if it's not already, but this
    # matrix won't get used if we pre-set the normalised slots.

    scrub = Scrublet(
        adata_obs.X,
        n_neighbors=n_neighbors,
        expected_doublet_rate=expected_doublet_rate,
        stdev_doublet_rate=stdev_doublet_rate,
        random_state=random_state,
    )

    # Ensure normalised matrix sparseness as Scrublet does
    # https://github.com/swolock/scrublet/blob/67f8ecbad14e8e1aa9c89b43dac6638cebe38640/src/scrublet/scrublet.py#L100

    scrub._counts_obs_norm = sparse.csc_matrix(adata_obs.X)
    scrub._counts_sim_norm = sparse.csc_matrix(adata_sim.X)

    scrub.doublet_parents_ = adata_sim.obsm["doublet_parents"]

    # Call scrublet-specific preprocessing where specified

    if mean_center and normalize_variance:
        pipeline.zscore(scrub)
    elif mean_center:
        pipeline.mean_center(scrub)
    elif normalize_variance:
        pipeline.normalize_variance(scrub)

    # Do PCA. Scrublet fits to the observed matrix and decomposes both observed
    # and simulated based on that fit, so we'll just let it do its thing rather
    # than trying to use Scanpy's PCA wrapper of the same functions.

    if mean_center:
        logg.info("Embedding transcriptomes using PCA...")
        pca_torch(scrub, n_prin_comps=n_prin_comps, random_state=scrub._random_state)
        #pipeline.pca(scrub, n_prin_comps=n_prin_comps, random_state=scrub._random_state)
    else:
        logg.info("Embedding transcriptomes using Truncated SVD...")
        pipeline.truncated_svd(
            scrub, n_prin_comps=n_prin_comps, random_state=scrub._random_state
        )

    # Score the doublets

    scrub.calculate_doublet_scores(
        use_approx_neighbors=use_approx_neighbors,
        distance_metric=knn_dist_metric,
        get_doublet_neighbor_parents=get_doublet_neighbor_parents,
    )

    # Actually call doublets

    scrub.call_doublets(threshold=threshold, verbose=verbose)

    # Store results in AnnData for return

    adata_obs.obs["doublet_score"] = scrub.doublet_scores_obs_

    # Store doublet Scrublet metadata

    adata_obs.uns["scrublet"] = {
        "doublet_scores_sim": scrub.doublet_scores_sim_,
        "doublet_parents": adata_sim.obsm["doublet_parents"],
        "parameters": {
            "expected_doublet_rate": expected_doublet_rate,
            "sim_doublet_ratio": (
                adata_sim.uns.get("scrublet", {})
                .get("parameters", {})
                .get("sim_doublet_ratio", None)
            ),
            "n_neighbors": n_neighbors,
            "random_state": random_state,
        },
    }

    # If threshold hasn't been located successfully then we couldn't make any
    # predictions. The user will get a warning from Scrublet, but we need to
    # set the boolean so that any downstream filtering on
    # predicted_doublet=False doesn't incorrectly filter cells. The user can
    # still use this object to generate the plot and derive a threshold
    # manually.

    if hasattr(scrub, "threshold_"):
        adata_obs.uns["scrublet"]["threshold"] = scrub.threshold_
        adata_obs.obs["predicted_doublet"] = scrub.predicted_doublets_
    else:
        adata_obs.obs["predicted_doublet"] = False

    if get_doublet_neighbor_parents:
        adata_obs.uns["scrublet"]["doublet_neighbor_parents"] = (
            scrub.doublet_neighbor_parents_
        )

    return adata_obs


@old_positionals(
    "layer", "sim_doublet_ratio", "synthetic_doublet_umi_subsampling", "random_seed"
)
def scrublet_simulate_doublets(
    adata: AnnData,
    *,
    layer: str | None = None,
    sim_doublet_ratio: float = 2.0,
    synthetic_doublet_umi_subsampling: float = 1.0,
    random_seed: _LegacyRandom = 0,
) -> AnnData:
    r"""Simulate doublets by adding counts of random observed transcriptome pairs.

    Generate synthetic doublets by randomly selecting pairs of observed cells
    and combining their transcriptomes to create artificial doublet profiles
    for training the doublet detection classifier.

    Arguments:
        adata: Annotated data matrix of shape n_obs × n_vars
        layer (str): Layer containing raw values, or None to use .X (default: None)
        sim_doublet_ratio (float): Number of doublets to simulate relative to observed cells (default: 2.0)
        synthetic_doublet_umi_subsampling (float): UMI sampling rate for doublet creation (default: 1.0)
        random_seed (int): Random seed for reproducible doublet simulation (default: 0)

    Returns:
        adata: AnnData object containing simulated doublets with metadata
    """
    X = _get_obs_rep(adata, layer=layer)
    scrub = Scrublet(X, random_state=random_seed)

    scrub.simulate_doublets(
        sim_doublet_ratio=sim_doublet_ratio,
        synthetic_doublet_umi_subsampling=synthetic_doublet_umi_subsampling,
    )

    adata_sim = AnnData(scrub._counts_sim)
    adata_sim.obs["n_counts"] = scrub._total_counts_sim
    adata_sim.obsm["doublet_parents"] = scrub.doublet_parents_
    adata_sim.uns["scrublet"] = {"parameters": {"sim_doublet_ratio": sim_doublet_ratio}}
    return adata_sim

def pca_torch(
    self: Scrublet,
    n_prin_comps: int = 50,
    *,
    random_state: _LegacyRandom = 0,
    svd_solver: Literal["auto", "full", "arpack", "randomized","gesvd", "gesvdj", "gesvda"] = "auto",
    use_gpu: bool = False,
) -> None:
    if self._counts_sim_norm is None:
        msg = "_counts_sim_norm is not set"
        raise RuntimeError(msg)
    

    X_obs = self._counts_obs_norm.toarray()
    X_sim = self._counts_sim_norm.toarray()

    if use_gpu:
        if svd_solver == "auto":
            svd_solver = "gesvd"
        try:
            from torchdr import PCA
        except ImportError:
            raise ImportError("torchdr is not installed. Please install it using `pip install torchdr`.")
        import torch
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        pca = PCA(
            n_components=n_prin_comps, random_state=None, svd_driver=svd_solver,
            device=device,
        ).fit(X_obs)
    else:
        from sklearn.decomposition import PCA
        if svd_solver == "auto":
            svd_solver = "arpack"
        pca = PCA(
            n_components=n_prin_comps, random_state=random_state, svd_solver=svd_solver,
        ).fit(X_obs)
    self.set_manifold(pca.transform(X_obs), pca.transform(X_sim))

